// distribution.cc
//
// Created 1998 by Jouni K Seppnen
//
// $Id: distribution.cc,v 1.13 1999/11/03 11:24:14 jks Exp $
//

#include <cctype>
#include <cstring>
using namespace std;

#include "common.h"
#include "distribution.h"
#include "main.h"

// The fixed table of defined distributions.
distributionDefinition::distDefVector *distributionDefinition::_allDistributions = 0; 
// The counter of generated code names
int distributionDefinition::nextCodename = 0;


void
distributionDefinition::initAllDistributions()
{
    if (_allDistributions != 0)
	return;			// already initialized

    // Create table
    _allDistributions = new distDefVector;

    // Fill in the table with the standard distributions
    _allDistributions->push_back(new distributionDefinition("beta", 2, "dBeta"));
    _allDistributions->push_back(new distributionDefinition("bin", 2, "dBin"));
    _allDistributions->push_back(new distributionDefinition("exp", 1, "dExp"));
    _allDistributions->push_back(new distributionDefinition("integer", 2, "dInteger"));
    _allDistributions->push_back(new distributionDefinition("gamma", 2, "dGamma"));
    _allDistributions->push_back(new distributionDefinition("geom", 1, "dGeom"));
    _allDistributions->push_back(new distributionDefinition("norm", 2, "dNorm"));
    _allDistributions->push_back(new distributionDefinition("poisson", 1, "dPoisson"));
    _allDistributions->push_back(new distributionDefinition("unif", 2, "dUnif"));
}

void 
distributionDefinition::newDistribution(const distributionDefinition *d)
{
    const_cast<vector<const distributionDefinition*>&>
	(allDistributions()).push_back(d);
}

void
distributionDefinition::printHeaderFile()
{
    distDefVector::const_iterator i;
    for (i = allDistributions().begin();
	 i != allDistributions().end();
	 i++)
	{
	    if ((*i)->isBuiltin())
		continue;	// don't need to print anything

	    FILE* f = c.file;
	    const distributionDefinition &d = *(*i);
	    const char* paramNames[26] = 
	    { "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", 
	      "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z" };
	    ASSERT(d.nParameters <= 26);

	    // Generate the strings of parameters that are used a few times below
	    string dclStr = "";	// declaration
	    string propDclStr = ""; // proposal declaration
	    string callStr = ""; // call

	    for (int j = 0; j < d.nParameters; j++) {
		dclStr += "double ";
		dclStr += paramNames[j]; callStr += paramNames[j];
		if (j+1 == d.nParameters) // just before the last comma
		    propDclStr = dclStr;
		dclStr += ", ";          callStr += ", ";
	    }
	    dclStr += "double value"; callStr += "value";
	    const char *dcl = dclStr.c_str(),
		       *propDcl = propDclStr.c_str(),
		       *call = callStr.c_str();

	    const char *code = d.codename->c_str(), 
		       *ucode = d.userCodename->c_str();

	    fprintf(f, "double %s(%s); /* must be supplied by user */\n", ucode, dcl);
	    fprintf(f, "class %s_class {\n", code);
	    fprintf(f, "public:\n");
	    fprintf(f, "    logprob density(%s)\n", dcl);
	    fprintf(f, "    { return %s(%s); }\n", ucode, call);
	    fprintf(f, "    realval propose(%s) { ASSERT(false); }\n", propDcl);
	    fprintf(f, "    logprob proposalDensity(%s)\n", dcl);
	    fprintf(f, "    { return density(%s); }\n", call);
	    fprintf(f, "};\n");
	    fprintf(f, "extern %s_class %s;\n\n", code, code);
	}
}

void 
distributionDefinition::printCodeFile()
{
    distDefVector::const_iterator i;
    for (i = allDistributions().begin();
	 i != allDistributions().end();
	 i++)
	{
	    if ((*i)->isBuiltin())
		continue;	// don't need to print anything
	    
	    FILE* f = c.file;
	    const distributionDefinition &d = *(*i);
	    const char *code = d.codename->c_str();

	    fprintf(f, "%s_class %s;\n", code, code);
	}
}


const distributionDefinition *
distributionDefinition::findDistribution(const string &name)
{
    // Compare caselessly
    string lc_name = name;
    for (int i = lc_name.length()-1; i>=0; i--) {
	lc_name.at(i) = tolower(lc_name.at(i));
    }

    for (distDefVector::const_iterator i = allDistributions().begin();
	 i != allDistributions().end();
	 i++) 
	{
	    if (lc_name.compare((*i)->giveName()) == 0) {
		return *i;
	    }
	}

    return 0;			// nothing found
}

distributionDefinition::distributionDefinition(
    const string &dname, int parms, const string &cname, bool standard
):
    name(dname), nParameters(parms), builtin(standard), 
    codename(0), userCodename(0)
{
    if (builtin) {
	codename = new string(cname);
    } else {
	userCodename = new string(cname);
	char buf[20];
	snprintf(buf, 20, "userDist%d", nextCodename++);
	codename = new string(buf);
    }
}


distributionRef::distributionRef(const string &dist, vector<expr*> *params):
    distribution(0), parameters(params)
{
    distribution = distributionDefinition::findDistribution(dist);

    if (distribution == 0) {
	c.parse_error("unknown distribution: %s", dist.c_str());

	string s = "*unknown: ";  s += dist; s += "*";
	string t = "/*unknown: "; t += dist; t += "*/";
	
	distribution = new distributionDefinition(s, 0, t);
    } else {
	if (parameters->size() != distribution->giveNParameters()) {
	    c.parse_error("distribution %s takes %d parameters, not %d",
			  distribution->giveName().c_str(), 
			  distribution->giveNParameters(), 
			  parameters->size());
	}
    }
    
}

void
distributionRef::findReferences()
{
    for_each(parameters->begin(), parameters->end(),
	     mem_fun(&expr::findReferences));
}

void
distributionRef::printCode()
{
    fprintf(c.file, "%s(", distribution->giveName().c_str());

    vector<expr*>::iterator
	i = parameters->begin(),
	end = parameters->end();
    for (; i != end; i++) {
	(*i)->printCode();
	if (i+1 != end)
	    fputs(", ", c.file);
    }

    fputs(")", c.file);
}

void
distributionRef::printClass()
{
    fputs(distribution->giveCodename().c_str(), c.file);
}

void
distributionRef::printParams()
{
    vector<expr*>::iterator
	i = parameters->begin(),
	end = parameters->end();
    for (; i != end; i++) {
	(*i)->printValue();
	if (i+1 != end)
	    fputs(", ", c.file);
    }
}


// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:
