// distribution.h
//
// Created 1998 by Jouni K Seppnen
//
// $Id: distribution.h,v 1.7 1999/11/01 11:35:26 jks Exp $
//

#ifndef DISTRIBUTION_H
#define DISTRIBUTION_H

#include <vector>
using namespace std;

#include "expr.h"

/** 
 * Encapsulates all we need to know about a distribution on the side of the
 * compiler.
 */
class distributionDefinition {
    typedef vector<const distributionDefinition*> distDefVector;
    /// Static table of distributions
    static distDefVector *_allDistributions;
    /// Function to initialize the allDistributions table
    static void initAllDistributions();
    /// Function that returns the list of distributions, making sure it exists
    static inline const distDefVector &allDistributions();

    /// Counter of generated code names
    static int nextCodename;

    /// Name of distribution, e.g. ``poisson''.
    const string name;
    /// How many parameters the density function takes.
    int nParameters;
    /** Whether this is a full, built-in distribution. Otherwise this
        is a user-defined distribution that only has the density function. */
    bool builtin;
    /** Name of the object that implements it in the Bassist library,
        e.g. ``dPoisson''. */
    const string *codename;
    /** In a user-defined distribution, name of the user-supplied
        density function. */
    const string *userCodename;
public:
    /// Constructor
    distributionDefinition(const string &dname, int parms, 
			   const string &cname, bool standard = true);
    /// Add new distribution
    static void newDistribution(const distributionDefinition *d);
    /// Returns a distribution with the given name, or null if none is found.
    static const distributionDefinition *findDistribution(const string &name);

    /// Print the header definitions for all user-defined distributions
    static void printHeaderFile();
    /// Print the necessary code for all user-defined distributions
    static void printCodeFile();

    /// Gives the name of this distribution.
    const string &giveName() const { return name; };
    /// Gives the number of parameters (the density function of) this distribution takes.
    int giveNParameters() const { return nParameters; };
    /// Is this a built-in distribution?
    bool isBuiltin() const { return builtin; };
    /// Gives the code name
    const string &giveCodename() const { return *codename; };
};

inline const distributionDefinition::distDefVector &
distributionDefinition::allDistributions()
{ 
    if (_allDistributions == 0) 
	initAllDistributions();
    return *_allDistributions;
}

/** 
 * A helper class for stochastic variables, proposal, and initialisation
 * distributions.
 */
class distributionRef {
    /// Which distribution is referenced by this reference
    const distributionDefinition *distribution;
    /// What parameters are used for the distribution by this reference
    vector<expr*> *parameters;
public:
    /// Constructor
    distributionRef(const string &dist, vector<expr*> *params);
    /// Accessor
    const distributionDefinition *giveDistribution() const { return distribution; };
    /// Accessor
    vector<expr*> *giveParameters() const { return parameters; };
    /// Resolve references recursively
    void findReferences();
    /// Print the code for this distribution reference
    void printCode();
    /// Print the class name of the referenced distribution
    void printClass();
    /// Print the parameters for this distribution reference
    void printParams();
};

#endif // DISTRIBUTION_H


// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:
