// expr.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: expr.h,v 1.11 1999/08/04 12:03:23 jks Exp $
//

#ifndef EXPR_H
#define EXPR_H

#include <string>
#include <set>
#include <vector>
using namespace std;

#include "node.h"
#include <stdio.h>

// Forward definitions
class variable;
class record;
class function;


/** Arithmetic operator type (+,-,*,/) needed by the arithmeticExpr
    class.
    @see arithmeticExpr
    @see arithmeticExpr::op */
enum arithmeticOp {
    /// Addition (+)
    PLUS,
    /// Subtraction (-)
    MINUS,
    /// Multiplication (*)
    MULT,
    /// Division (/) 
    DIV
};


/** The base class of all the mathematical expressions in the parse tree.

    An expr is a mathematical expression, which may have several
    subexpressions. A mathematical expression in the model is
    converted to a tree of exprs, which reform the expression when
    evaluated from left to right. */
class expr : public parseTreeNode {
public:
    /** Prints the model specification to the standard output.

	This function has no default implementation in expr class. 
	@see node::printCode
	@see compiler::printTree */
    virtual void printCode() = 0;

    /** Prints the C++ code name (C++ code) of the expression.
	This function is used by variable references.

	Default behaviour is to call printCode.
	@see expr::printCode
	@see node::printCode */
    virtual inline void printCodeName() { printCode(); }

    /// Prints the expression as C++ code. No default.
    virtual void printValue() = 0;

    /** Tests if this is a constant expression. Default is false.
	@return true if this is a constant expression */
    virtual inline bool isConst() { return false; }
    
    /** Tests if this expression should be expressed as an implicit
	composite in the generated code. Default is false.
	@return true if this is an implicit composite */
    virtual inline bool isComposite() { return false; }

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. 

	Default implementation is to do nothing, because only the
	variableRef class has something to do.
	@see compiler::parseInput
	@see variableRef::findReferences */
    virtual inline void findReferences() {}
    
    /** Collects all the variable references from this expression to the
	list, which is needed during the generation of the dependency
	function of a deterministic variable. 

	Default implementation is to do nothing, because only the
	variableRef class has something to do.
	@see determVar::genDepFunction
	@see variableRef::collectVars */
    virtual inline void collectVars(vector<string> *) {}

    /// Virtual destructor since there are other virtual functions
    virtual ~expr() {}
};



/// The parse tree numerical constant class.
class constant : public expr {
protected:
    /// The value of the constant.
    double val;

public:  
    /** The constructor.
	@param d The value of the constant. */
    constant(double d) : val(d) {};

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();
    
    /** Prints the expression as C++ code. Corresponds to the printing
	of the model specification. */
    inline void printValue() { printCode(); }

    /** Tests if this is a constant expression. This is.
	@return true */
    inline bool isConst() { return true; }
};



/** The parse tree arithmetic expression class.

    This class represents all the two-component arithmetic expressions
    in the parse tree, for example 2 + 3. The operations are defined
    by arithmeticOp, and at the moment, they consist of the basic
    operations +, -, * and /. The operands are other expressions.
    @see arithmeticOp */
class arithmeticExpr : public expr {
protected:
    /// Expression for the left side of the operation. 
    expr *l;
    /// Expression for the right side of the operation. 
    expr *r;
    /// Arithmetic operation type (+,-,*,/).
    arithmeticOp op;

public:  
    /** The constructor.
	@param L Expression for the left side of the operation. 
	@param o Arithmetic operation type (+,-,*,/).
	@param R Expression for the right side of the operation.   */	
    arithmeticExpr(expr *L, arithmeticOp o, expr *R);
    
    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();
    
    /// Prints the expression as C++ code.
    void printValue();
    
    /** Tests if this is a constant expression. The answer is asked
	recursively from the subexpressions.

	@return true if this is a constant expression */
    inline bool isConst() { return (l->isConst() && r->isConst()); }
    
    /** Tests if this expression should be expressed as an implicit
	composite in the generated code. This should.
	@return true */
    inline bool isComposite() { return true; }
    
    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. 

	The operation is passed on recursively to the subexpressions.
	@see compiler::parseInput */
    inline void findReferences() { l->findReferences(); r->findReferences(); }

    /** Collects all the variable references from this expression to the
	list. This is needed to generate the dependency function of a
	deterministic variable. 

	The operation is passed on recursively to the subexpressions.
	@see determVar::genDepFunction */
    inline void collectVars(vector<string> *list) { l->collectVars(list); r->collectVars(list); }
};



/** The parse tree negation expression class.

    This class represents all the unary negation expressions (-x) in
    the parse tree. */
class negationExpr : public expr {
protected:
    /// The expression to be transformed to its negation.
    expr *r;
    
public:  
    /** The constructor.
	@param R The expression to be transformed to its negation. */	
    negationExpr(expr *R);
    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();

    /// Prints the expression as C++ code.
    void printValue();

    /** Tests if this is a constant expression. The answer is asked
	recursively from the expression to be transformed to its negation.
	@return true if this is a constant expression */
    inline bool isConst() { return r->isConst(); }

    /** Tests if this expression should be expressed as an implicit
	composite in the generated code. This should.
	@return true  */
    inline bool isComposite() { return true; }

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. 

	The operation is passed on recursively to the expression to
	negate. 
	@see compiler::parseInput */
    inline void findReferences() { r->findReferences(); }

    /** Collects all the variable references from this expression to the
	list. This is needed to generate the dependency function of a
	deterministic variable. 

	The operation is passed on recursively to the expression to
	negate. 
	@see determVar::genDepFunction */
    inline void collectVars(vector<string> *list) { r->collectVars(list); }
};



/** A parse tree class for the expressions inside parenthesis.

    This class represents all the parenthesis expressions, for example
    (2*3). When the expression is generated in C++, the evaluation
    order cannot be changed and the parenthesis cannot be ignored,
    because the RPN calculation order cannot be expressed in C++. */
class parenExpr : public expr {
protected:
    /// The expression inside the parenthesis.
    expr *e;

public:
    /** The constructor.
	@param E The expression inside the parenthesis. */	
    parenExpr(expr *E);

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();
    
    /// Prints the expression as C++ code.
    void printValue();

    /** Tests if this is a constant expression. The answer is asked
	recursively from the expression inside the parenthesis.
	@return true if this is a constant expression */
    inline bool isConst() { return e->isConst(); }

    /** Tests if this expression should be expressed as an implicit
	composite in the generated code. 

	The operation is passed on recursively to the the expression
	inside the parenthesis. 
	@return true if this is an implicit composite */
    inline bool isComposite() { return e->isComposite(); }

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. 

	The operation is passed on recursively to the the expression
	inside the parenthesis. 
	@see compiler::parseInput */
    inline void findReferences() { e->findReferences(); }

    /** Collects all the variable references from this expression to the
	list. This is needed to generate the dependency function of a
	deterministic variable. 

	The operation is passed on recursively to the expression
	inside the parenthesis.
	@see determVar::genDepFunction */
    inline void collectVars(vector<string> *list) { e->collectVars(list); }
};


/** The parse tree variable reference class.

    All the variable references in the model like 'alpha' or
    'rat.alpha' are instances of this class.  */
class variableRef: public expr {
    /// The name of the referred variable.
    string varName;
    /// The possible record part of the variable name that is referred to.
    string recName;
    /** The name of the model specification file in which this
	reference was used. Needed when reporting reference errors.
	@see findReferences
	@see compiler::reference_error */
    string dFile;
    /** The line number of the model specification file in which this
	reference was used. Needed when reporting reference errors.
	@see findReferences
	@see compiler::reference_error */
    int line;
    
    /// The referred parse tree variable.
    variable *var;
    /// The possible record in which the referred variable is defined.
    variable *rec;
    /** Indicates if the referred variable is a local (not global) 
    	variable. */
    bool local;
    /// The C++ reference name for the referred variable. 
    string wholeName;
    /// The record in which this reference was defined, if any.
    record *myRecord;
    /** Whether to use oldValue instead of the actual value in the code.
        Used in the proposal distribution arguments. */
    bool useOldValue;
    
public:  
    /** The constructor for the simple variable references.
	@param v The name of the referred variable. */	
    variableRef(const string &v);

    /** The constructor for the record variable references.
	@param v The variable part of the reference name.
	@param r The record part of the reference name. */	
    variableRef(const string &r, const string &v);

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();

    /** Prints the C++ code name of the expression.
	The C++ code names differ from variable names: all the C++
	code names begin with 'var\_' to avoid name conflicts between the 
	simulator program variables and the model variables. */
    void printCodeName();

    /// Prints the expression as C++ code.
    void printValue();

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. This function searches for the referred variable, 
	which should be found; if it is not found, an error occurs.
	@see compiler::parseInput
	@see compiler::reference_error */
    void findReferences();

    /** Constructs the C++ equivalent for the original variable name,
	sets the value of the 'wholename' attribute, and adds the
	generated code to the given list. The list is needed during
	the generation of the dependency function of a deterministic
	variable.
	@param l The list into which the name should be added.
	@see variableRef::wholename
	@see determVar::genDepFunction */
    void collectVars(vector<string> *l);
};


/** A parse tree class for function call prototypes.

    This class represents all the function calls "f(x)" in the parse
    tree. */
class functionCall : public expr {
    /// A pointer to the function definion.
    function *func;
    /// List of the parameters in the function call. 
    vector<expr*> *params;
    
public:  

    /** The constructor.
	@param f Name of the function.
	@param l List of the parameters of the function call.  */	
    functionCall(const string &f, vector<expr*> *l);

    /** Tests if this expression should be expressed as an implicit
	composite in the generated code. A function call should be.
	@return true */
    inline bool isComposite() { return true; }

    /** Prints the model specification to the standard output. The
	operation is passed on recursively to the parameters of the
	function call. 
	@see node::printCode
	@see compiler::printTree */
    void printCode();

    /** Prints the expression as C++ code. The operation is passed on
	recursively to the parameters of the function call. */
    void printValue();

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. 

	The operation is passed on recursively to the parameters of
	the function call.
	@see compiler::parseInput */
    void findReferences();

    /** Collects all the variable references from this expression to the
	list. This is needed to generate the dependency function of a
	deterministic variable. 

	The operation is passed on recursively to the parameters of
	the function call.
	@see determVar::genDepFunction */
    void collectVars(vector<string> *l);
};

/** A parse tree class for aggregate function calls. */

class aggregateFunctionCall : public expr {
protected:
    /// Which aggregate function gets called.
    string aggregateFunctionName;

    /// The name of the record part of the function call argument.
    string argumentRecordName;
    /// The resolved record part of the function call argument.
    record *argumentRecord;

    /// The name of the variable part of the function call argument. 
    string argumentVariableName;
    /// The resolved variable part of the function call argument.
    variable *argumentVariable;

    /// Whether the return value is constant.
    bool constant;

    /// The file in which this call appeared.
    string dFile;
    /// The line on which this call appeared.
    int dLine;
    
public:
    /// Constructor.
    aggregateFunctionCall (const string &function, 
			   const string &record,
			   const string &variable = "");
    
    /// 
    void printCode();
    ///
    void printCodeName() { printCode(); /* ??? FIXME */ };
    ///
    void printValue();


    ///
    static void printImplementationPrototype(string function, variable* var);
    ///
    static void printImplementationCode(string function, variable* var);

    /** Whether this expression should be expressed as an implicit
        composite in the generated code. This one should. */
    inline bool isComposite() { return true; }
    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. */
    void findReferences();
    ///
    void collectVars(vector<string> *l);
    /// Whether this expression is constant. This one may or may not be.
    virtual bool isConst() { return constant; }


};

#endif // EXPR_H

//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// c-basic-offset: 4           //
// End:                        //
//-----------------------------//
