// composite.h
//
// Created 1997 by Karri-Pekka Laakso
//
// $Id: composite.h,v 1.7 1999/08/10 08:29:25 jks Exp $
//

#ifndef COMPOSITE_H
#define COMPOSITE_H

#include <vector>
using namespace std;

#include "quantity.h"

// Forward declaration
class parameter;
class scanner;

/** The class of the deterministic nodes of the model. 

*/
class composite : public quantity {
protected:
    /** The list of all the composites that are defined in the model
	specification, i.e. the ones that possibly should be printed
	during the simulation. The implicitly generated composites,
	that have no corresponding deterministic node in the
	simulation, are excluded from this list. */
    static vector<composite*> composites;
    
public:
    /** The constructor.
	@param name The name of this composite
	@param implicit Indicates if the composite is an implicitly
	generated composite */
    composite(const string &name, bool implicit=false) : quantity(name) {
	if (! implicit)
	    composites.push_back(this);
    }

    /** The dependents of this composite call this function to tell,
	that they depend on this composites value. The implementation
	is undefined here, because it depends solely on the expression
	that defines this composites value. */
    virtual void addDependent(parameter *) = 0;	

    /** Gives the value of this composite.  The implementation
	is undefined here, because it depends solely on the expression
	that defines this composites value. */
    virtual inline double value() const = 0;

    /** Checks that this deterministic variable is ready for
	simulation. The implementation is empty, because there is
	nothing to do. */
    virtual void initialize() { }
    
    /** This function may assign this composite a data value or an
	initial value, if the string "name = value" is found from the
	input. Because the values of composites are defined by their
	expressions, by default askValue does not assign any value to
	the composite. This can be overriden in a subclass.
       	@return false */
    virtual bool askValue(scanner &, const char *, bool) { return false; }

    /** Returns an iterator to the beginning of the vector of all the
	composites defined in the model. 
	@see composites 
	@return The iterator 
	*/
    static vector<composite*>::const_iterator begin() {
	return composites.begin();
    }

    /** Returns an iterator to the end of the vector of all the composites
        defined in the model.
      @see composites
      @see begin
      @return The iterator
      */
    static vector<composite*>::const_iterator end() {
	return composites.end();
    }
};

#endif // COMPOSITE_H

// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:

