// distributions.cc
//
// Created 1997 by Karri-Pekka Laakso
//
// $Id: distributions.cc,v 1.12 2000/01/17 14:04:58 jks Exp $
//

#include "distributions.h"
#include "constant.h"
#include <limits.h>

//----------------------------------------
// Global distribution objects
//----------------------------------------

dbin dBin;
dnorm dNorm;
dgamma dGamma;
dbeta dBeta;
dexp dExp;
dpoisson dPoisson;
dgeom dGeom;
dunif dUnif;
dinteger dInteger;


// --------------------------------------------------------------------
// lngamma
//
// Compute the gamma function. Since gamma gets rather large rather
// fast, we really compute the logarithm.
//
// If you need the factorial, it's probably a bad idea to use
// exp(lngamma(n+1.0)).
//
// We use the Lanczos approximation with $\gamma=5$, $N=6$.
// --------------------------------------------------------------------

double
lngamma(double x)
{
    DEBUG('d', "lngamma(%G)\n", x);
    const double EPSILON = 0.000001; // FIXME: Chosen arbitrarily.
    const double LNSQRT2PI = (M_LN2 + log(M_PI))/2;
    const double
	C0 =   1.000000000178,	// These magic numbers come directly from
	C1 =  76.180091729406,	// Lanczos's paper. We could probably 
	C2 = -86.505320327112,	// get a bit better accuracy by doing
	C3 =  24.014098222230,	// the math ourselves, but I don't know
	C4 =  -1.231739516140,	// whether it's worth anything.
	C5 =   0.001208580030,	// Lanczos, C.: A Precision Approximation
	C6 =  -0.000005363820;	// of the Gamma Function, J. SIAM Numer. Anal.
				// Ser. B, Vol 1, pp. 86-96

    // Gamma function is negative in some argument realms, and we are
    // computing the logarithm -> negative gamma is not supported
    ASSERT(x > 0.0);
    
    if (x>1) {
	// This is the main algorithm. It's actually generalisable to complex
	// arguments whose real part is greater than 1.
    
	return
	    (x-0.5) * log(x+4.5) - x - 4.5 + LNSQRT2PI +
	    log(C0 + C1/(x) + C2/(x+1) + C3/(x+2) + C4/(x+3) + C5/(x+4) + C6/(x+5));

    } else if (1 - x < EPSILON) {
	// Gamma(1) = 1
	return 0.0;
    } else {
	// For 0 < x < 1 we use $\Gamma(z+1) = z\Gamma(z)$
	return lngamma(1+x) - log(x);
    }
}




//----------------------------------------
// Binomial distribution
//----------------------------------------


logprob **dbin::factors = 0;


//----------------------------------------------------------------------
// dbin::dbin 
//  Initialize the binomial distribution 
// 
// N - amount of Bernoulli trials
// P - trial succes propability
// --------------------------------------------------------------------

dbin::dbin() { 
    DEBUG('d', "Initializing dbin distribution\n");

    ASSERT(factors == 0);	// This bites, if there are multiple
				// instances of this class  
    factors = new logprob*[DBIN_MAX_N];
    for (int i = 0; i < DBIN_MAX_N; i++)
	factors[i] = 0;
    
    savedVal = savedP = savedN = savedProb = 0;
}



//----------------------------------------------------------------------
// dbin::binomialFactor
//  count the logarithm of the binomial factor, log (n! / k!(n-k)!)
// 
// n, k - binomial factor parametres
// --------------------------------------------------------------------

logprob
dbin::binomialFactor(long n, long k) {
    DEBUG('d', "binomialFactor(%ld, %ld)\n", n, k);
    double upper = 0, lower = 0;
    long i, ulimit, llimit;

    ASSERT(n >= 1);
    ASSERT(n >= k);

    // Basically the binomial factor is (n! / k!(n-k)!)
    // However, that is a rather inefficient way to calculate it. The
    // first idea is to split the calculation in two parts:
    //     n! / k! == k+1 * ... * n (the upper part) and
    //    (n-k)!                   (the lower part

    // This is still inefficient, because (k+1 * ... * n) and (n-k)!
    // contain several identical terms, which will be simplified in
    // the division. Therefore, when k < (n-1 / 2), whe consider the
    // binomial as n(k) (= n-k+1 *...* n) and k!, and when k is
    // larger or equal, as (k+1 * .. n) and (n-k)!.

    // The integer division cuts off the desimal part, therefore
    // (n-1 / 2) has to be written as (n+1 / 2) to find the right
    // value of k.
    
    if (k < ((n + 1) / 2)) {
	llimit = k;
	ulimit = n-k+1;
    } else {
	llimit = n-k;
	ulimit = k+1;
    }
    
    // We want the logarithm as the result, so instead of
    // log(x*x-1*...) we do log(x) + log (x-1) + ...
    for (i = n; i >= ulimit; i--) {  
	upper += log((double)i);
    }


    for (i = 1; i <= llimit; i++) {
	lower += log((double)i);
    }

    // log(x/y) = log(x) - log(y)
    return (upper - lower);
}


//----------------------------------------------------------------------
// dbin::density
//  The logarithm of the binomial density function
//        (  n  )
//    log (value) + value * log (p / 1 - p) + n * log (1 - p) 
//
// value - integer number of trials
// --------------------------------------------------------------------

logprob
dbin::density(intval N, realval p, intval value) {
    long val = (long) value;
    long n = (long) N;

    DEBUG('d', "binomial density for %lu; N=%lu, p=%G\n", val, n, p);

    // Special cases
    if (n > DBIN_MAX_N)
	return dPoisson.density(N*p, value);

    ASSERT(p >= 0.0 && p <= 1.0 && n >= 0);
    
    if (val > n || val < 0)
	return -BASS_INFINITY;		// Out of range, probability = 0

    if (n == 0)				// val = 0, n = 0,	
	return 0;			// probability = 1

    if (p == 0.0) {			// Prob is 1 for val = 0, zero     
	if (val == 0)			// otherwise
	    return 0;
	else
	    return -BASS_INFINITY;
    }
    
    if (p == 1.0) {			// prob is 1 for val = n, zero     
	if (val == n)			// otherwise
	    return 0;
	else
	    return -BASS_INFINITY;
    }
    
    ASSERT(p > 0.0 && p < 1.0);		// Equality on either side yields log(0) :(

    if (value == savedVal && 		// Speed up things, in case we
	p == savedP && n == savedN)	// have just counted this
	return savedProb;		// density (in this case we
    // _can_ compare float equality).  

    savedP = p;
    savedN = n;
    savedVal = value;
    
    //  We can speed things up a lot by using a table for binomial
    //  coefficients. A static table of all such tables exists, and if
    //  there is no table for this constant n, it is created and
    //  initialized.

    logprob *factorTable = factors[n-1];
    // Note: n = 0 was handled already

    if (factorTable == 0) { 	// Init factorTable for this N 
	DEBUG('d', "Initializing factor table for %lu\n", n);
	factorTable = new logprob[n];
    	for (long i = 1; i <= n; i++) 
	    factorTable[i-1] = binomialFactor(n, i);
	factors[n-1] = factorTable;
    }
    
    savedProb = factorTable[val-1] + val * (log(p) - log(1-p)) + n * log(1-p);

    DEBUG('d', "Counted density using table: %G\n", savedProb);
    
    return savedProb;
}



//----------------------------------------------------------------------
// dbin::propose
//  Propose a new value from the distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------

intval
dbin::propose(intval N, realval p) {
    double r = randH(0, 1);
    long newValue = 0;
    double sumD = 0;
    long n = (long) N;
    DEBUG('d', "binomial proposal; N=%lu, p=%G\n", n, p);
    
    if (p < 0.0 || p > 1.0 || n < 1) {
	DEBUG('d', "out of range: proposing 0\n");
	return 0;
    }

    if (n > DBIN_MAX_N) {
	DEBUG('d', "using Poisson approximation\n");
	return dPoisson.propose(N*p);
    }
    
    for (; newValue < n; newValue++) {
	sumD += exp(density(N, p, (double) newValue));

	if (r <= sumD)
	    break;
    }

    ASSERT(newValue >= 0 && newValue <= N);

    DEBUG('d', "proposing %lu\n", newValue);
    return (intval) newValue;    
}



//----------------------------------------------------------------------
// dbin::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------

logprob
dbin::proposalDensity(intval N, realval p, intval newValue) {
    if ((long) N > DBIN_MAX_N)
	return dPoisson.density(N*p, newValue);
    
    return density(N, p, newValue);
}





//----------------------------------------
// Normal distribution
//----------------------------------------


//----------------------------------------------------------------------
// dnorm::dnorm
//  Initialize the normal distribution
//
// myy - expectation
// sigma2 - variance
// --------------------------------------------------------------------

dnorm::dnorm() {
    DEBUG('d', "Initializing dnorm distribution\n");
    have_saved_variate = 0;
}



//----------------------------------------------------------------------
// dnorm::standard_normal_density
//  The logarithm of the standard normal density function
//
// value - value for which the density in std normal distribution is to be
// evaluated, value in R
// --------------------------------------------------------------------

logprob 
dnorm::standard_normal_density(realval value)
{
    const double LN_SQRT_1_2PI = log(sqrt(1/(2*M_PI)));
    double result = LN_SQRT_1_2PI - 0.5*value*value;

    DEBUG('d', "std normal density at %G: %G\n", value, result);
    return result;
}     



//----------------------------------------------------------------------
// dnorm::density
//  The logarithm of the normal density function
// 
// value - value for which the density in normal distribution is to be
// evaluated, value in R
// --------------------------------------------------------------------

logprob
dnorm::density(realval myy, realval sigma2, realval value) {
    DEBUG('d', "normal density for %G; mu=%G, sigma2=%G\n", value, myy, sigma2);
    ASSERT(sigma2 > 0);

    realval sigma = sqrt(sigma2);
    return -log(sigma) + standard_normal_density((value-myy)/sigma);
}



//----------------------------------------------------------------------
// dnorm::propose
//  Propose a new value from the distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------

realval
dnorm::propose(realval myy, realval sigma2) {
    DEBUG('d', "normal proposal; mu=%G, sigma2=%G\n", myy, sigma2);
    if (sigma2 <= 0)
	return 0; 	// Out of range, give an arbitrary proposal

    // The Box-Muller method
    if (have_saved_variate) {
	have_saved_variate = 0;
	DEBUG('d', "using saved variate: %G\n", myy + saved_variate*sqrt(sigma2));
	return myy + saved_variate*sqrt(sigma2);
    } else {
	double x1, x2, sum_sq, factor;
      
      // Generate a point (x1,x2) within the unit circle.
	do {
	    x1 = randH(-1.0, 1.0);
	    x2 = randH(-1.0, 1.0);
	    sum_sq = x1*x1 + x2*x2;
	} while (sum_sq > 1.0 || sum_sq == 0.0);

	factor = sqrt(-2.0 * log(sum_sq)/sum_sq);
	saved_variate = factor*x1; have_saved_variate = 1;
	DEBUG('d', "using new variate: %G\n", myy + factor*x2*sqrt(sigma2));
	return myy + factor*x2*sqrt(sigma2);
    }
} 



//----------------------------------------------------------------------
// dnorm::proposalDensity
//  Gives the density of the proposed value
//
// 
// --------------------------------------------------------------------

logprob
dnorm::proposalDensity(realval myy, realval sigma2, realval newValue) {
    return density(myy, sigma2, newValue);
} 


//----------------------------------------
// Gamma distribution
//----------------------------------------


//----------------------------------------------------------------------
// dgamma::dgamma
//  Initialize the gamma distribution
//
// 
// --------------------------------------------------------------------

dgamma::dgamma() {
    DEBUG('d', "Initializing dgamma distribution\n");
}


//----------------------------------------------------------------------
// dgamma::density
//  The logarithm of the exponential density function
//
// 
// alpha,lambda - expectation = alpha/lambda, variance = alpha/lambda*lambda
// value - value for which the density in gamma distribution is to be
// evaluated, value in R
// --------------------------------------------------------------------

logprob
dgamma::density ( realval alpha, realval lambda, realval value)
{
    DEBUG('d', "gamma density for %G; alpha=%G, lambda=%G\n", value, alpha, lambda);
    ASSERT(alpha > 0 && lambda > 0);

#if 0 // the assert takes care of this
    if (lambda <= 0) {
	DEBUG('d', "out of range: zero density\n");
	return -BASS_INFINITY;
    }
#endif

    if (value > 0) {
	return (alpha * log (lambda) + (alpha - 1) * log (value) 
		- lambda * value - lngamma(alpha));
    } else {
	DEBUG('d', "out of range: zero density\n");
	return -BASS_INFINITY;
    }
}



//----------------------------------------------------------------------
// dgamma::propose
//  Propose a new value from gamma distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// Algorithm used is Best's rejection algorithm XG (1978)
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------

realval
dgamma::propose (realval alpha, realval lambda)
{
    DEBUG('d', "gamma proposal; alpha=%G, lambda=%G\n", alpha, lambda);
    // density function: f(x,a,b)=b^a * x^(a-1) * e^(-bx)/gamma(a) 

    double x;  // we will return x / lambda

    if (alpha <= 1.0)
    {
	if (alpha == 1.0)
	    x = dExp.propose(1); 
	else {
	    DEBUG('d', "Using Ahrens-Dieter algorithm GS\n");
	    bool done = false;

	    while(!done) {
		// Step 1
		double u = randH(0.0, 1.0);
		double b = (M_E + alpha)/M_E;
		double p = b*u;
		DEBUG('d', "Step 1: p=%G\n", p);
		if (p <= 1.0) {
		    // Step 2: Case x <= 1
		    x = pow(p, 1.0/alpha);
		    double u_star = randH(0.0, 1.0);
		    if (u_star > exp(-x)) {
			DEBUG('d', "Step 2: rejecting, x=%G, u*=%G\n", x, u_star);
		    } else {
			DEBUG('d', "Step 2: accepting, x=%G, u*=%G\n", x, u_star);
			done = true;
		    }
		} else { 
		    // Step 3: Case x > 1
		    x = -log( (b-p)/alpha );
		    double u_star = randH(0.0, 1.0);
		    if (u_star > pow(x, alpha-1)) {
			DEBUG('d', "Step 3: rejecting, x=%G, u*=%G\n", x, u_star);
		    } else {
			DEBUG('d', "Step 3: accepting, x=%G, u*=%G\n", x, u_star);
			done = true;
		    }
		}
	    }
	}
    } 
    else			// alpha > 1.0
    {
	DEBUG('d', "Using old algorithm for case alpha > 1.0\n");
	double b, c, u, v, w, y, z;
	int accept = 0;

	b = alpha - 1.0;
	c = 3 * alpha - 0.75;
	do
	{
	    u = randH (0.0, 1.0);
	    v = randH (0.0, 1.0);
	    w = u * (1 - u);
	    y = sqrt (c / w) * (u -0.5);
	    x = b + y;
	    if (x >= 0)
	    {
		z = 64 * pow (w, 3) * pow (v, 2);
		if ((x != 0) && (z <= 1 - (2 * pow (y, 2) / x)))
		    accept = 1;
		if (accept == 0)
		{
		    if ((z > 0) && (x != 0) &&
			(log (z) <= 2 * (b * log (x / b) - y)))
			accept = 1;
		}
	    }
	}
	while (accept == 0);
    }                      /* x ~ gamma (alpha,1) => x/lambda ~ gamma (alpha, lambda)  */
    return (x / lambda);
}     



//----------------------------------------------------------------------
// dgamma::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.             
//  
// --------------------------------------------------------------------
           
logprob
dgamma::proposalDensity(realval alpha, realval lambda, realval newValue)
{
    return density(alpha,lambda,newValue);
}
             
//----------------------------------------
// Beta distribution
//----------------------------------------

//---------------------------------------------------------------------
// dbeta::dbeta
//  
//
// 
// --------------------------------------------------------------------


dbeta::dbeta()
{
}


//---------------------------------------------------------------------
// dbeta::density
//  
//
// 
// --------------------------------------------------------------------

logprob
dbeta::density(realval alpha, realval beta, realval value)
{
    DEBUG('d', "beta density for %G; alpha=%G, beta=%G\n", value, alpha, beta);
    ASSERT(alpha > 0 && beta > 0);

    if (value >= 0.0 && value <= 1.0) {
	return
	    lngamma(alpha + beta) - lngamma(alpha) - lngamma(beta)
	    + (alpha - 1.0) * log(value) + (beta - 1.0) * log(1.0 - value);
    } else {
	DEBUG('d', "out of range [0.0, 1.0]: zero density\n");
	return -BASS_INFINITY;
    }
}


//---------------------------------------------------------------------
// dbeta::propose
//  
//
// 
// --------------------------------------------------------------------

realval
dbeta::propose(realval alpha, realval beta)
{
    DEBUG('d', "beta proposal; alpha=%G, beta=%G\n", alpha, beta);
    // Cheng, R. C. H.: Generating Beta Variates with Nonintegral 
    // Shape Parameters, Comm. ACM, Apr 1987, Vol. 21, Nr 4

    // Note: the meanings of a and alpha, and of b and beta, are
    // reversed compared to the paper. We also denote gamma with c
    // and delta with d for the sake of consistency.

    if (alpha <= 0 || beta <= 0) {
	DEBUG('d', "out of range, proposing 0\n");
	return 0;
    }

    const double LN4 = log(4.0), LN5PLUS1 = log(5.0) + 1.0;
    double X;			// The final result

    if(min(alpha, beta) > 1) {
	DEBUG('d', "using Cheng's BB algorithm\n");

	int ab_swapped;		// we want alpha <= beta; have we swapped?
	double a, b, c, U1, U2, V, W, Z, R, S, T;

	// Initialisation
	if (alpha > beta) {
	    double temp = beta;
	    beta = alpha;
	    alpha = temp;

	    ab_swapped = 1;
	} else {
	    ab_swapped = 0;
	}
	a = alpha + beta;
	b = sqrt((a - 2)/(2*alpha*beta - a));
	c = alpha + 1/b;

	do {
	    // Step 1
	    do {
		U1 = randH(0.0, 1.0);
	    } while (U1 == 0.0 || U1 == 1.0);
	    do {
		U2 = randH(0.0, 1.0);
	    } while (U2 == 0.0 || U2 == 1.0);
	    V = b * log(U1/(1 - U1));
	    W = alpha * exp(V);
	    Z = U1*U1*U2;
	    R = c * V - LN4;
	    S = alpha + R - W;

	    // Step 2
	    if (S + LN5PLUS1 >= 5*Z)
		break;

	    // Step 3
	    T = log(Z);
	    if (S >= T)
		break;

	    // Step 4
	} while (R + a*log(a/(beta + W)) < T);

	// Step 5
	if (ab_swapped) {
	    X = beta/(beta + W);
	} else {
	    X = W/(beta + W);
	}

    } else {			// min(alpha, beta) <= 1
	DEBUG('d', "Using Cheng's BC algorithm\n");

	int ab_swapped;		// we want alpha >= beta, REVERSED wrt BB
	double a, b, d, k1, k2, U1, U2, Y, Z, V=0, W=0;
    
	// Initialisation
	if (alpha < beta) {
	    double temp = beta;
	    beta = alpha;
	    alpha = temp;
	    
	    ab_swapped = 1;
	} else {
	    ab_swapped = 0;
	}
	a = alpha + beta;
	b = 1.0/beta;
	d = 1.0 + alpha - beta;
	k1 = d*(1.0/(4*18) + 3.0/(4*18)*beta)/(alpha*b - 14.0/18);
	k2 = 0.25 + (0.5 + 0.25/d)*beta;
	
	// Now, the BC algorithm. First, I tried to convert it to
	// a more structural notation, but introduced subtle bugs.
	// This version is more faithful to Cheng's presentation.

    BCstep1:
	do {
	    U1 = randH(0.0, 1.0);
	} while (U1 == 0.0 || U1 == 1.0);
	do {
	    U2 = randH(0.0, 1.0);
	} while (U2 == 0.0 || U2 == 1.0);

	if (U1 >= 0.5) {
	    goto BCstep3;
	}

	// BCstep2:
	Y = U1*U2;
	Z = U1*Y;
	if (0.25*U2 + Z - Y >= k1) {
	    goto BCstep1;
	} else {
	    goto BCstep5;
	}

    BCstep3:
	Z = U1*U1*U2;
	if (Z <= 0.25) {
	    V = b*log(U1/(1 - U1));
	    W = alpha*exp(V);
	    goto BCstep6;
	}

	// BCstep4:
	if (Z >= k2) {
	    goto BCstep1;
	}

    BCstep5:
	V = b*log(U1/(1 - U1));
	W = alpha*exp(V);
	if (a*(log(a/(beta + W)) + V) - LN4 < log(Z)) {
	    goto BCstep1;
	}

    BCstep6:
	if (ab_swapped) {
	    X = beta/(beta + W);
	} else {
	    X = W/(beta + W);
	}
    }
    
    return X;
}


//---------------------------------------------------------------------
// dbeta::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.             
//  
// --------------------------------------------------------------------
           
logprob 
dbeta::proposalDensity(realval alpha, realval beta, realval newValue)
{
    return density(alpha, beta, newValue);
}




//----------------------------------------
// Exponential distribution
//----------------------------------------



//----------------------------------------------------------------------
// dexp::dexp
//  Initialize the exponential distribution
//
// 
// --------------------------------------------------------------------

dexp::dexp() {
    DEBUG('d', "Initializing dexp distribution\n");
}


                
//----------------------------------------------------------------------
// dexp::density
//  The logarithm of the exponential density function
//
// lambda - expectation = 1/lambda, variance = 1/lambda*lambda
// value - value for which the density in exp distribution is to be
// evaluated, value in R
// --------------------------------------------------------------------

logprob
dexp::density (realval lambda, realval value)
{
    DEBUG('d', "exponential density for %G; lambda=%G\n", value, lambda);
    ASSERT(lambda > 0);

    if (value < 0.0) {
	DEBUG('d', "out of range: zero density\n");
	return -BASS_INFINITY;
    }
    return log(lambda) - (lambda)*value;
}      



//----------------------------------------------------------------------
// dexp::propose
//  Propose a new value from exponential distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------
         
realval
dexp::propose(realval lambda)
{
    DEBUG('d', "exponential proposal; lambda=%G\n", lambda);

    double y;

    if (lambda <= 0) {
	DEBUG('d', "out of range, proposing 0\n");
	return 0;
    }
    
    do 
	y=randH(0.0,1.0);
    while (y == 1.0);

    return (-log(1-y)/lambda); 
}
           



//----------------------------------------------------------------------
// dexp::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.             
//  
// --------------------------------------------------------------------


logprob
dexp::proposalDensity(realval l, realval newValue)
{        
    return density(l, newValue);
}



//----------------------------------------
// Poisson distribution
//----------------------------------------


//----------------------------------------------------------------------
// dpoisson::dpoisson
//  Initialize the Poisson distribution
//
// 
// --------------------------------------------------------------------


dpoisson::dpoisson() {
    DEBUG('d', "Initializing dpoisson distribution\n");
    savedK = -1;
    savedL = -1.0;
}


                
//----------------------------------------------------------------------
// dpoisson::density
//  The logarithm of the Poisson density function
// 
// k - value for which the density in Poisson distribution is to be
// evaluated, k in N
// --------------------------------------------------------------------


logprob
dpoisson::density (realval lambda, intval value)
{
    DEBUG('d', "Poisson density for %G; lambda=%G\n", value, lambda);
    ASSERT(lambda > 0);

    if (value < 0.0) {
	DEBUG('d', "out of range: zero density\n");
	return -BASS_INFINITY;
    }
    
    long k = (long) value;
    
    // Was this just calculated?
    if (k == savedK && lambda == savedL) {
	DEBUG('d', "using cached result %G\n", savedProb);
	return savedProb;
    }
    
    savedK = k;
    savedL = lambda;
    savedProb = -lambda + k*log(lambda) - lngamma(k+1);

    DEBUG('d', "result is %G\n", savedProb);
    return savedProb;
}


//---------------------------------------------------------------------
// dpoisson::propose
//  Propose a new value from poisson distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------

intval
dpoisson::propose(realval lambda)
{
    DEBUG('d', "Poisson proposal; lambda=%G\n", lambda);
    if (lambda <= 0.0) {
	DEBUG('d', "out of range: zero density\n");
	return 0;
    }
    
    double r = randH(0, 1);
    double sumD, p;

    sumD = p = exp(-lambda);
    long i = 0;
    
    for (; r > sumD && i < LONG_MAX; i++) { 
	if (p > 0)
	    p *= lambda / (i+1);
	else  // p might be 0, if lambda is big enough
	    p = lambda / (i+1);
	sumD += p;
    }

    DEBUG('d', "i=%ld after loop\n", i);
    ASSERT(i >= 0 && i <= LONG_MAX);
    
    return (intval) i;
}


//----------------------------------------------------------------------
// dpoisson::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.             
//  
// --------------------------------------------------------------------


logprob
dpoisson::proposalDensity(realval lambda, intval newValue)
{        
    return density(lambda, newValue);
}




//----------------------------------------
// Geometric distribution
//----------------------------------------


//----------------------------------------------------------------------
// dgeom::dgeom
//  Initialize the geometric distribution
//
// --------------------------------------------------------------------


dgeom::dgeom() {
    DEBUG('d', "Initializing dgeom distribution\n");
}



//----------------------------------------------------------------------
// dgeom::density
//  The logarithm of the geometric density function
//
//
// p - expectation = (1-p)/p, variance = (1-p)/p*p
// value - value for which the density in geometric distribution is to be
// evaluated, value in N
// --------------------------------------------------------------------


logprob
dgeom::density (realval p, intval value)
{
    DEBUG('d', "geometric density for %G; p=%G\n", value, p);
    ASSERT(p > 0.0 && p < 1.0);

    long v = (long) value;

    if (v < 1) {
	DEBUG('d', "out of range: zero density\n");
	return -BASS_INFINITY;
    }
    
    return ((v-1)*log(1-p) + log(p));
}


//---------------------------------------------------------------------
// dgeom::propose
//  Propose a new value from geometric distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------

intval
dgeom::propose(realval p)
{
    DEBUG('d', "geometric proposal; p=%G\n", p);
    if (p <= 0.0 || p >= 1.0) {
	DEBUG('d', "out of range: proposing 1\n");
	return 1;
    }

    double r = randH(0, 1);
    double sumD, prob;
    long i=1;
    sumD = prob = p;
    
    for (; r > sumD && i < LONG_MAX; i++) { 
	prob *= (1-p);
	sumD += prob;
    }

    DEBUG('d', "i=%ld after loop\n", i);
    ASSERT(i >= 0 && i <= LONG_MAX);
    
    return (intval) i;
}


//----------------------------------------------------------------------
// dgeom::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.             
//  
// --------------------------------------------------------------------


logprob
dgeom::proposalDensity(realval p, intval newValue) {        
    return density(p, newValue);
}




//----------------------------------------
// Uniform distribution
//----------------------------------------


//----------------------------------------------------------------------
// dunif::dunif
//  Initialize the uniform distribution
//
// --------------------------------------------------------------------


dunif::dunif() {
    DEBUG('d', " Initializing dunif distribution\n");
}



//----------------------------------------------------------------------
// dunif::density
//  The logarithm of the uniform density function
//
//
// a,b - start and end points of the density function (density 1/b-a)
// value - value for which the density in uniform distribution is to be
// evaluated, value in R
// --------------------------------------------------------------------


logprob
dunif::density(realval start, realval end, realval value) {
    DEBUG('d', "uniform density for %G; start=%G, end=%G\n", value, start, end);
    ASSERT(start < end);

    if (start!=end)
	if ((value<start) || (value>end)) return -BASS_INFINITY;
	else return log(1/(end-start));

    else if (value==start) return 0; else return -BASS_INFINITY;

}



//----------------------------------------------------------------------
// dunif::propose
//  Propose a new value from the distribution for MH simulation
//  The proposition is done by random, and it's propability follows the
//  distribution.
//
// In this case, the old value is not needed for the proposal.
// --------------------------------------------------------------------


realval
dunif::propose(realval start, realval end) {
    DEBUG('d', "uniform proposal; start=%G, end=%G\n", start, end);
    if (start > end)
	return 0;

    return randH(start,end);
}
            


//----------------------------------------------------------------------
// dunif::proposalDensity
//  The density of a proposed value for the MH simulation
//
// newValue - the value for which we want the propability
// In this case, the old value is not needed for the proposal.             
//  
// --------------------------------------------------------------------


logprob
dunif::proposalDensity(realval a, realval b, realval newValue) {
    return density(a,b,newValue);
}     


//----------------------------------------------------------------------
// Uniform integer distribution
//----------------------------------------------------------------------

dinteger::dinteger()
{
    DEBUG('d', "Initializing dinteger distribution\n");
}

logprob
dinteger::density(intval low, intval high, intval value)
{
    DEBUG('d', "uniform-integer density for %G; low=%G, high=%G\n", value, low, high);
    ASSERT(low <= high);

    if (low <= value && value <= high) {
	long a = static_cast<long>(low);
	long b = static_cast<long>(high);
	long n = b - a + 1;

	return static_cast<logprob>(log(1.0/n));
    } else {
	return -BASS_INFINITY;
    }
}

intval
dinteger::propose(intval low, intval high)
{
    DEBUG('d', "uniform-integer proposal; low=%G, high=%G\n", low, high);

    if (low > high)
	return low;

    // Let's just hope low and high are representable integers.
    // (FIXME if you feel like redoing parts of Bassist using longs.)

    double r = floor(randH(low, high+1.0));
    ASSERT(low <= r && r <= high);

    return static_cast<intval>(r);
}

logprob
dinteger::proposalDensity(intval low, intval high, intval newValue)
{
    return density(low, high, newValue);
}


// Local Variables:
// mode: c++
// c-basic-offset: 4
// case-replace: nil
// make-backup-files: t
// End:

