// distributions.h 
// 
// Created 1997 by Karri-Pekka Laakso
//
// @(#)$Id: distributions.h,v 1.3 1999/08/03 14:58:32 jks Exp $
//

#ifndef DISTRIBUTIONS_H
#define DISTRIBUTIONS_H

#include <math.h>
#include <stdio.h>

#include "common.h"
#include "quantity.h"



double lngamma(double x);
// returns the natural logarithm of gamma of x


//----------------------------------------
// Binomial distribution 
//----------------------------------------

#define DBIN_MAX_N 100	// Maximum n for binomial distribution. With
			// an N bigger than this, Poisson
			// approximation Bin(n, p) ~ Poisson(np) is used.


/**
   Binomial distribution is an asymmetric discrete probability 
   distribution. It is said that random variable \TEX{$\Theta$}
   is distributed according to binomial distribution
   with parameters n and p 
   (\TEX{$\Theta \sim bin(n,p)$}),
   if its density function has a form 
   \TEX{$f(x) = {n \choose k} p^{k}(1-p)^{n-k}$}. 
   Expectation of the binomial distribution
   is 
   \TEX{$np$} and variance is 
   \TEX{$np(1-p)$}.
*/
class dbin {
private:
/// Previous value that was used in calculation of density.
    realval savedVal;
/// Previous probability that was used in calculation of density.
    realval savedP;
/// Previous number of trials that was used in calculation of density.
    realval savedN;	// Optimisation; don't count
/** 
	Previous density. This attribute (and attributes 
	savedVal, savedP, savedN) are saved in order
	to get more efficient calculation.
*/
    logprob savedProb;			// the same density twice. 
    
protected:
///
    logprob *factorTable;

/**
    	Calculates logarithm of binomial factor n over k.
    	\TEX{$log\Big({n \choose k}\Big) = log\Big(\frac{n!}{k!(n-k)!}\Big)=log\Big(\frac{(k+1) \cdots n}{(n-k)!}\Big)=\sum_{i=k+1}^{n} log(i) - \sum_{j=1}^{n-k} log(j)$}
	@param n 
	@param k
    	@return logarithm of binomial factor n over k 
*/
    logprob binomialFactor(long n, long k);

/**
	Table that contains previously calculated values of 
	the binomial coefficients. Therefore
	each binomial coefficient is calculated at most
	once during simulation.
*/
    static logprob **factors;

public:
/**
	Constructor. Initializes protected and private data structures and 
	global table of binomial factors.
*/
    dbin();

/**
    	Calculates the logarithm of the binomial density function
    	@param N The number of Bernoulli trials
    	@param p Probability of success
    	@param value Number of successes in N Bernoulli trials
    	@return The logarithm of the binomial density function
*/ 
    virtual logprob density(intval N, realval p, intval value);
/**
	Returns a random value from the binomial distribution.
	Distribution of the proposals follows
	distribution \TEX{$bin(N,p)$}. 
	@param N The number of Bernoulli trials
	@param p Probability of success
	@return A new proposal for the simulation
*/
    virtual intval propose(intval N, realval p);
    
/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dBin.
	@param N Number of Bernoulli trials
	@param p Probability of success
	@param newValue New proposed value (see method propose)
	@param Logarithm of the density of the proposed value
*/
    virtual logprob proposalDensity(intval N, realval p, intval newValue);
};




//----------------------------------------
// Normal distribution
//----------------------------------------

/** 
	Normal distribution is a well-known symmetric continous
	probability distribution.
	It is said, that random variable \TEX{$\Theta$} is normally distributed
	with parameters \TEX{$\mu$} and \TEX{$\sigma$}
	(\TEX{$\Theta \sim N(\mu, \sigma^{2})$}), 
	if its density function has a form 
	\TEX{$f(x) = \frac{exp\Big(-0.5(\frac{x-\mu}
	{\sigma})^{2}\Big)}{\sqrt{2\pi}\sigma}$}. Mean of the
	normal distribution is \TEX{$\mu$} and variance is 
	\TEX{$\sigma^{2}$}.
*/

class dnorm {
private:
/**
	Logarithm of density function of the distribution \TEX{$N(0,1)$}.
	@param value Value for which the density in standard normal
	distribution is to be evaluated.
	@return Logarithm of the density at point value.
*/
    logprob standard_normal_density(double value);

  // The algorithm generates two independent normal variates at a
  // time; we will keep one and return the other, and on the next call
  // only return the one we saved.

/** 	
	Integer have_saved_variate indicates whether we have
	a value that is generated during a previous call. If we 
	have one, we return it, otherwise we generate new two 
	normal variates.
	
	Values of a given normal distribution are generated as follows.
	Firstly, we generate two independent normal variates at time.
	We will keep one and return the other, and on the next 
	call we return the one we saved.

*/
  int have_saved_variate;

/**
	Value of the saved normal variate.
*/
  realval saved_variate;

public:
/**
	Constructor. Initializes attribute have saved variate.
*/
    dnorm();

/**
	Calculates logarithm of the normal density function.
	@param myy Mean parameter 
	@param sigma2 Variance parameter
	@param value value for which the density in N(myy, sigma2)
	is to be evaluated.
	@return Logarithm of the normal density function
	at point value.
*/
    logprob density(realval myy, realval sigma2, realval value);

/**
	Returns a random value from the normal distribution N(myy, sigma2).
	This method uses the Box-Muller method. See
	"Box, G.E.P, M.E. Muller 1958; A note on the generation of 
	random normal deviates, Annals Math. Stat, V. 29, pp. 610-611". 	
	@param myy Mean of the normal distribution
	@param sigma2 Variance of the normal distribution
	@return A new proposal for simulation
*/
    realval propose(realval myy, realval sigma2);
/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dNorm.
	@param myy Mean parameter for the normal distribution.
	@param sigma2 Variance parameter for the normal distribution.
	@param newValue Proposed value
	@return Logarithm of the density of the proposed value.
*/
    logprob proposalDensity(realval myy, realval sigma2, realval newValue);
};


//----------------------------------------
// Gamma distribution
//----------------------------------------

/**
	Gamma distribution is an asymmetric continous probability distribution. 
	It is said, that random variable \TEX{$\Theta$} is distributed according 
	to gamma distribution with parameters \TEX{$\alpha$} and 
	\TEX{$\lambda$} (\TEX{$\Theta \sim Gamma(\alpha,\lambda)$}),
	if its density function has a form 
	\TEX{$f(x) = \frac{\lambda e^{-\lambda x} (\lambda x)^{\alpha - 1}} {\Gamma(\alpha)}$, $x \geq 0$}, where 
	\TEX{$\Gamma(\alpha) = \int_{0}^{\infty}e^{-y} y^{\alpha - 1}dy = (\alpha - 1) \Gamma (\alpha - 1)$}. 
	Consequently,
	\TEX{$\Gamma(n) = (n-1)!$ for all $n \in {\cal N}$}.
	Expectation of the gamma distribution is 
	\TEX{$E(X) = \frac{\alpha}{\lambda}$} and
	variance is \TEX{$V(X) = \frac{\alpha}{\lambda^{2}}$}. 
*/
class dgamma {
private:
/**
	Generates random variables from \TEX{$Gamma(a,1)$}, \TEX{$a<1$}.
	@param d First parameter for \TEX{$Gamma(a,1)$}.
	@return Logarithm of density of the gamma distribution
	at point a.
*/
    logprob rand_gamma_below(double d);
    
public:
/// Constructor
    dgamma();
/**
	Calculates logarithm of the density function of the gamma
	distribution
	@param alpha First parameter of the gamma distribution
	@param lambda Second parameter of the gamma distribution
	@param value Value for which the density in gamma distribution
	is to evaluated. 
	@return Logarithm of the density function of 
	distribution \TEX{$Gamma(alpha, lambda)$} at point value.
*/
    logprob density(realval alpha, realval lambda, realval value);

/**
	Returns a random value from the gamma distribution. Algorithm
	is implemented according to Best's rejection
	algorithm XG (1978).
	@param alpha First parameter of the gamma distribution
	@param lambda Second parameter of the gamma distribution
	@return A new proposal for simulation
*/
    realval propose(realval alpha, realval lambda);

/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dGamma.
	@param alpha First parameter of the gamma distribution
	@param lambda Second parameter of the gamma distribution
	@param value Value for which we want the density
	@return Logarithm of the density of the proposed value
*/
    logprob proposalDensity(realval alpha, realval lambda, realval newValue);
};


//----------------------------------------
// Beta distribution
//----------------------------------------

/**
	Beta distribution is an asymmetric continous probability distribution. 
	It said that random variable
	\TEX{$\Theta$} is distributed according beta distribution with
	parameters \TEX{$\alpha$} and \TEX{$\beta$} 
	(\TEX{$\Theta \sim Beta(\alpha, \beta)$},
	\TEX{$\alpha, \beta > 0$}), if its density function has a form
	\TEX{$f(x) = \frac{\Gamma(\alpha + \beta)}{\Gamma(\alpha)\Gamma(\beta)} (1-x)^{(\beta - 1)}x^{(\alpha - 1)}, x \in ]0,1[$} 
	Expectation of the beta distribution is 
	\TEX{$E(X) = \frac{\alpha}{\alpha+\beta}$}
	and variance is 
	\TEX{$V(X) = \frac{\alpha \beta}{(\alpha+\beta)^{2}(1+\alpha+\beta)}$}
  */
class dbeta {
public:
/// Constructor
    dbeta();
/**
	Logarithm of density function of the beta distribution.
	@param alpha The first parameter of the beta distribution
	@param beta The second parameter of the beta distribution
	@param value Value for which the density in beta distribution
	is to be evaluated.
	@return Logarithm of density of the distribution 
	beta(alpha, beta) at point value.
*/
    logprob density(realval alpha, realval beta, realval value);

/**
	Returns a random value from beta distribution.
    	The proposals follows distribution $beta(alpha, beta)$.
    	Algorithm is implemented according
    	to an article written by Cheng, R. C. H.	
    	Cheng, R. C. H.: Generating Beta Variates with Nonintegral
        Shape Parameters, Comm. ACM, Apr 1987, Vol. 21, Nr 4      
	@param alpha First paramer of the beta distribution
	@param beta Second parameter of the beta distribution
	@return A new value for simulation.
*/
    realval propose(realval alpha, realval beta);
/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dBeta.
	@param alpha The first parameter of the beta distribution
	@param beta The second parameter of the beta distribution
	@param newValue New proposed value for this dBeta
	@return
*/
    logprob proposalDensity(realval alpha, realval beta, realval newValue);
};



//----------------------------------------
// Exponential distribution
//----------------------------------------

/**
	Exponential distribution is an asymmetric continous probability
	distribution. It is said that random 
	variable \TEX{$\Theta$} is distributed according to 
	exponential distribtution with parameter
	\TEX{$\lambda$} (\TEX{$\Theta \sim exp(\lambda)$}), 
	if its density function has a form
	\TEX{$f(x) = \lambda e^{-\lambda x}$, $\lambda > 0$},
	\TEX{$x \geq 0$}. Mean of the exponential distribution
	is \TEX{$E(X) = \mu =  \frac{1}{\lambda}$} and variance is
	\TEX{$V(X) = \mu^{2} = \frac{1}{\lambda^{2}}$}.
*/
class dexp {
public:
/// Constructor
    dexp();
/**
	Logarithm of the exponential density function.
	@param lambda Parameter of the exponential distribution.
	@param value Value for which logarithm of density of the exponential
	distribution is to be evaluated.
	@return Logarithm of the density of exponential distribution
	at point value.
*/

    logprob density(realval lambda, realval value);

/**
	Returns a random value from the exponential distribution. Proposals
	follows distribution exp(lambda).
	@param lambda Parameter of exponential distribution
	@return Proposal for a new value
*/
    realval propose(realval lambda);

/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dExp.
	@param lambda Parameter of exponential distribution
	@param newValue The value for which we the
	probability.
	@return Logarithm of density of the proposed value.
*/
    logprob proposalDensity(realval lambda, realval newValue);  
};



//----------------------------------------
// Poisson distribution
//----------------------------------------

/**
	Poisson distribution is a discrete probability distribution. 
	It is said that a random 
	variable \TEX{$\Theta$} is distributed according to
	the Poisson distribution with parameter \TEX{$\lambda$},
	if its density function has a form
	\TEX{$f(x) = \frac{\lambda^{x} e^{-\lambda}}{x!}$}, where
	\TEX{$x \in \{0,1,2,\ldots\}$}.
  */
class dpoisson {
/// Previously used value 
    long savedK;
/// Previously used value
    realval savedL;
/// Previous density (for optimization details).
    realval savedProb;
    
public:
/// Constructor
    dpoisson();
/**
	Logarithm of the Poisson density function.
	@param lambda Parameter of the Poisson distribution
	@param value Value for which the density in 
	Poisson distribution is to be evaluated.
	@return
*/
    logprob density(realval lambda, intval value);
/**
	Returns a random value from the poisson distribution. Proposals
	follows distribution Poisson(lambda). 
	@param lambda Parameter of the Poisson distribution
	@param value The value for which we want
	the probability.
	@return Logarithm of the density of the Poisson distribution
	at the point value.
*/
    intval propose(realval lambda);
/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dPoisson.
	@param lambda Parameter of the Poisson distribution
	@param newValue Proposed value
	@return Probability of proposing newValue.
*/
    logprob proposalDensity(realval lambda, intval newValue);  
};



//----------------------------------------
// Geometric distribution
//----------------------------------------

/**
	Geometric distribution is an asymmetric discrete probability
	distribution. It is said that random variable
	\TEX{$\Theta$} is distributed according to geometrical
	distribution with parameter \TEX{$p$} (\TEX{$\Theta \sim Geom(p)$}), if its 
	density function has the form \TEX{$f(x) = p(1-p)^{x}$},
	\TEX{$x \in {1, 2, \ldots}$}. Mean of the geometric 
	distribution is \TEX{$E(X) = \frac{1-p}{p}$} and
	variance is \TEX{$V(X) = \frac{1-p}{p^{2}}$}.
  */
class dgeom {
public:
/// Constructor
    dgeom();
/**
	The logarithmic density of the geometric distribution.
	@param Parameter of the geometrical distribution
	@param value Value for which the logarithmic density
	in geometric distribution is to evaluated.
	@return Logarithm of the density of the
	distribution Geom(p) at point value.
*/
    logprob density(realval p, intval value);

/**
	Returns a random value from the geometric distribution.
	Proposals follows distribution geom(p).
	@param Parameter of the geometric distribution
	@return A new proposal for the parameter of the
	geometric distribution.
*/
    intval propose(realval p);

/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dGeom.
	@param Parameter of the geometric distribution
	@param newValue Value for which we want the
	density.
	@return Logarithm of the density of the proposed
	value.
*/
    logprob proposalDensity(realval p, intval newValue);  
};




//----------------------------------------
// Uniform distribution
//----------------------------------------

/**
	Uniform distribution is a symmetric continous probability distribution.
	It said that random variable \TEX{$\Theta$} is distributed
	according to uniform distribution with parameters
	\TEX{$a$} and \TEX{$b$} (\TEX{$\Theta \sim unif(a,b)$}), 
	if its density function has a form
	\TEX{$f(x) = \frac{1}{b-a}$}, if \TEX{$x \in [a,b]$} and 0 otherwise.
*/
class dunif {
public:

/// Constructor. 
    dunif();
/**
	Logarithm of the uniform distribution.
	@param start First parameter of the uniform distribution.
	@param end Second parameter of the uniform distribution.
	@param value Value for which the density in uniform
	distribution is to be evaluated.
	@return
*/
    logprob density(realval start, realval end, realval value);
/**
	Returns a random value from the uniform distribution. Proposals
	follows distribution Unif(start, end).
	@param start First parameter of the uniform distribution.
	@param end Second parameter of the uniform distribution.
	@return Proposal
*/
    realval propose(realval start, realval end);
/**
	Logarithm of the proposal density from 
	the currentValue to the new proposed value in the Markov Chain
	that is determined by this dUnif.
	@param start The first parameter of the uniform distribution.
	@param end The second parameter of the uniform distribution.
	@return The density of the uniform distribution
	at the proposed value (newValue).
*/
    logprob proposalDensity(realval start, realval end, realval newValue);
};           


//----------------------------------------------------------------------
// Uniform integer distribution
//----------------------------------------------------------------------

/** The uniform integer distribution is a discrete version of the
    ordinary uniform distribution. Each integer from low to high (inclusive)
    is equally probable. NOTE: Behaviour is undefined if low and high
    are not representable integers.
*/

class dinteger {
public:
  /// Constructor
  dinteger();
  /** Logarithm of the density function for the uniform integer distribution.
    @param low The lower limit of the distribution
    @param high The upper limit of the distribution
    @param value Value at which the density function is to be evaluated
    @return The logarithm of the density function
    */
  logprob density(intval low, intval high, intval value);
  /** Returns a random value from the uniform integer distribution.
    Each integer from low to high (inclusive) is equally probable.
    @param low The lower limit of the distribution
    @param high The upper limit of the distribution
    @return An integer in [low,high]
   */
  intval propose(intval low, intval high);
  /** Logarithm of the proposal density function.
    This simply calls density.
    @see density
    */
  logprob proposalDensity(intval low, intval high, intval newValue);
};


//----------------------------------------
// Global instances of distributions
//----------------------------------------

extern dbin dBin;
extern dnorm dNorm;
extern dgamma dGamma;
extern dbeta dBeta;
extern dexp dExp;
extern dpoisson dPoisson;
extern dgeom dGeom;
extern dunif dUnif;
extern dinteger dInteger;


#endif // DISTRIBUTIONS_H

//------------------------------------------------
// Local Variables:				//
// mode: c++					//
// make-backup-files: t				//
// End:						//
//------------------------------------------------
