// parameter.cc
//
// Created 1997 by Karri-Pekka Laakso
//
// $Id: parameter.cc,v 1.14 1999/11/09 12:36:49 jks Exp $

#include <vector>
#include <algorithm>
using namespace std;

#include "parameter.h"
#include "scanner.h"
#include "main.h"

extern simulator *s;

vector<parameter*> parameter::parameters;

//---------------------------------------------------------------------
// parameter::parameter
//  The parameter constructors. 
//
// 
// --------------------------------------------------------------------

parameter::parameter(const string &n):
    quantity(n), currentValue(1), dependents(0), data(false), inited(false)
{ 
    DEBUG('p', " Constructing parameter %s\n", n.c_str());

    parameters.push_back(this);
}


//----------------------------------------------------------------------
// parameter::askValue
//  Assigns this parameter a value and sets it as data, if the
//  string "name = value" is found from the input.
//
// The value is of type double. Returns true if the sequence was
// found, false if it was not found.
// --------------------------------------------------------------------
// 

bool
parameter::askValue(scanner &sc, const char *n, bool init)
{
    // Get the next identifier, but don't move input, in case it
    // wasn't the one we're looking for
    char *nextId = sc.next(false);
    
    if (n == 0)
	n = name.c_str();
    
    if (nextId != 0 && strcmp(n, nextId) == 0) {
	// It matched, so first eat the id name
	sc.next();
	sc.get("=");
	if ( !init)
	    // "i" means initial value in data file, not data
	    init = sc.ask("i"); 
	double value = sc.getDoubleValue();
	if (init)
	    this->initialize(value);
	else
	    this->setData(value);
	return true;
    }

    return false;
}


//----------------------------------------------------------------------
// parameter::addDependent
//  Adds a new parameter to the dependency list.
//
// 
// --------------------------------------------------------------------

void
parameter::addDependent(parameter *p) {
    if (dependents == 0) 
	dependents = new vector<parameter*>;

    if (find(dependents->begin(), dependents->end(), p) == dependents->end())
	dependents->push_back(p);
}


//----------------------------------------------------------------------
// parameter::giveBeginIterator
//  Default iterator
//
// 
// --------------------------------------------------------------------

inline vector<parameter*>::const_iterator 
parameter::giveBeginIterator() const {
    return dependents->begin();
}

//----------------------------------------------------------------------
// parameter::giveEndIterator
//  Default iterator
//
// 
// --------------------------------------------------------------------

inline vector<parameter*>::const_iterator 
parameter::giveEndIterator() const {
    return dependents->end();
}


//---------------------------------------------------------------------
// parameter::MH
//  The Metropolis-Hastings simulation algorithm
//
// --------------------------------------------------------------------

bool
parameter::MH() 
{
    DEBUG('m', "Metropolis-Hastings for %s\n", identifier().c_str());

    if (data) {
	DEBUG('M', "- Fixed data");
	return true;
    }

    double oldValue = currentValue;
    double newValue = propose(currentValue);
    
    // Old value calculations
    
    logprob oldProbability = density();
    logprob backwardTransitionProb = proposalDensity(newValue, oldValue);
    DEBUG('M', "- old value %G, own old prob: %G, backward trans p: %G\n",
	  currentValue, oldProbability, backwardTransitionProb);

    if (oldProbability <= -BASS_INFINITY) {
	DEBUG('o', "MH for %s: oldProbability <= -BASS_INFINITY\n", identifier().c_str());
	DEBUG('o', "- old value %G, own old prob: %G, backward trans p: %G\n",
	      oldValue, oldProbability, backwardTransitionProb);
    } else {
	// no use computing dependents, if value itself is impossible
	parameter *p;
	if (dependents)
	    for (vector<parameter*>::const_iterator i = dependents->begin(); 
		 i != dependents->end(); 
		 ++i) 
		{
		    p = *i;
		    DEBUG('M', "  o dep %s, p=%G\n", p->identifier().c_str(), p->density());
		    oldProbability += p->density();
		}
    }

    DEBUG('m', "*** old (%G), p=%G\n ", oldValue, oldProbability);

    // New value calculations

    currentValue = newValue;
    logprob newProbability = density();
    logprob forwardTransitionProb = proposalDensity(oldValue, newValue);
    DEBUG('M', "- new value %G, own new prob: %G, forward trans p: %G\n", 
	  newValue, newProbability, forwardTransitionProb);

    if (newProbability <= -BASS_INFINITY) {
	DEBUG('o', "MH for %s: newProbability <= -BASS_INFINITY\n", identifier().c_str());
	DEBUG('o', "- new value %G, own new prob: %G, forward trans p: %G\n", 
	      newValue, newProbability, forwardTransitionProb);
    } else {
	// no use computing dependents, if value itself is impossible
	parameter *p;
	if (dependents)
	    for (vector<parameter*>::const_iterator i = dependents->begin();
		 i != dependents->end();
		 ++i)
		{
		    p = *i;
		    DEBUG('M', "  o dep %s, p=%G\n", p->identifier().c_str(), p->density());
		    newProbability += p->density();
		}
    }

    DEBUG('m', "*** new (%G), p=%G\n", newValue, newProbability);

    // Metropolis-Hastings decision
    
    double ratio = 
	(newProbability + backwardTransitionProb)
	-
	(oldProbability + forwardTransitionProb);
    DEBUG('m', "*** log ratio: %G, exp(log ratio): %G\n", ratio, exp(ratio));

    double result = true;
    if (oldProbability <= -BASS_INFINITY && newProbability <= -BASS_INFINITY) {
	DEBUG('o', "MH rejecting due to infeasible state\n");
	result = false;
    }
    
    if (newProbability > -BASS_INFINITY) {
	if (ratio >= 0) {
	    DEBUG('m', "*** log ratio >= 0, accepting immediately\n");
	} else {
	    double threshold = randH(0.0, 1.0);
	    DEBUG('m', "*** random threshold: %G\n", threshold);
	    if (exp(ratio) > threshold) {
		// accept (do nothing)
		DEBUG('m', "===> new value accepted\n");
	    } else {
		// do not accept
		currentValue = oldValue;
		DEBUG('m', "===> old value restored (didn't pass threshold)\n");
	    }
	}
    } else {
	// do not accept, restore old value
	currentValue = oldValue;
	DEBUG('m', "===> old value restored (right away)\n");
    }
    
    return result;
}



//---------------------------------------------------------------------
// parameter::setData
//  Provides a way to change this parameter to data.
//
// --------------------------------------------------------------------


void
parameter::setData(double value) {
    data = true;
    currentValue = value;
    inited = true;
    
    // parameters is the list of all simulatable parameters, and now
    // the parameter is becoming data.
    vector<parameter*>::iterator myPlace =
	find(parameters.begin(), parameters.end(), this);
    ASSERT(myPlace != parameters.end());
    parameters.erase(myPlace);
}


//---------------------------------------------------------------------
// parameter::~parameter
//  De-allocate a parameter. Remove it from the parameter list, and
//  deallocate it's own lists.
//
// 
// --------------------------------------------------------------------

parameter::~parameter() {
    remove(parameters.begin(), parameters.end(), this);

    if (dependents != 0) {
	delete dependents;
	dependents = 0;
    }
    
    return;
}



///////////////////////////////////////////////////////////////////////
// cluster experiment/HT


//---------------------------------------------------------------------
// cluster::printValue
//  
//
// 
// --------------------------------------------------------------------

void
cluster::printValue(FILE *file) {
    vector<parameter*>::const_iterator i = begin();
    parameter *p;

    while (i != end()) {
	p = *i++;
	fprintf(file, "\n %s : %G", p->identifier().c_str(), p->value());
    }
}



//---------------------------------------------------------------------
// cluster::MH
//  A modified Metropolis-Hastings for clusters of variables
//
// --------------------------------------------------------------------


bool
cluster::MH() 
{
    DEBUG('m', "Metropolis-Hastings for %s: ", identifier().c_str());

    // for each component
    //   propose
    // compute total likelihood
    // decide to accept of reject

    // generic variables:
    parameter *p;
    vector<parameter *>::const_iterator it;

    // 1. compute current partial likelihood
    logprob oldProbability = 0;
    // 1a. first for components
    it = begin();
    while (it != end()) {
	oldProbability += (*it++)->density();
    }

    // 1b. then for dependents
    if (oldProbability > -BASS_INFINITY) {
	// no use computing dependents, if value itself is impossible
	it = dependents->begin();
	while (it != dependents->end()) {
	    oldProbability += (*it++)->density();
	}
    }

    // 2. propose new values
    it = begin();
    while (it != end()) {
	(*it++)->proposeAndSetValue();
    }

    // 3. compute new partial likelihood
    logprob newProbability = 0;
    // 3a. first for components
    it = begin();
    while (it != end()) {
	newProbability += (*it++)->density();
    }

    // 3b. then for dependents
    if (newProbability > -BASS_INFINITY) {
	// no use computing dependents, if value itself is impossible
	vector<parameter*>::const_iterator it = giveBeginIterator();
	while (it != giveEndIterator()) {
	    newProbability += (*it++)->density();
	}
    }

    // 4. compute proposal probabilities
    logprob proposalDensityForOld = 0;
    logprob proposalDensityForNew = 0;
    it = begin();
    while (it != end()) {
	p = *it++;
	proposalDensityForOld += p->proposalDensityBackward();
	proposalDensityForNew += p->proposalDensityForward();
    }

    ASSERT(oldProbability > -BASS_INFINITY || newProbability > -BASS_INFINITY);

    if ((newProbability > -BASS_INFINITY) &&
	((newProbability + proposalDensityForOld >=
	  oldProbability + proposalDensityForNew) ||
	 (exp((newProbability + proposalDensityForOld) -
	      (oldProbability + proposalDensityForNew))
	  > randH(0.0,1.0)))) {
	// accept (do nothing)
	DEBUG('C', "- new values accepted\n");
    } else {
	// do not accept, restore old values
	it = begin();
	while (it != end()) {
	    (*it++)->restoreOldValue();
	}
	DEBUG('C', "- old values restored\n");
    }

    return true;
}


// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:
