// parameter.h
//
// Created 1997 by Karri-Pekka Laakso
//
// @(#)$Id: parameter.h,v 1.13 1999/08/10 08:29:25 jks Exp $
//

#ifndef PARAMETER_H
#define PARAMETER_H

#include <vector>
using namespace std;

#include <stdlib.h>
#include <stdio.h>
#include "common.h"
#include "distributions.h"
#include "quantity.h"

/**
The simulatable quantities of the model are instances of class parameter.
A parameter has a distribution and a proposer for MH-simulation.
All the densities and probabilities are represented as logarithms.

Parameter is the only subclass of quantity (except generated
class composite) that actually implements
dependency mechanism. Parameter has a list (dependents) that
consists of parameters that are its dependents. Therefore this
list is called dependency list. Dependency list is used
as follows: First, when this parameter is being initialized
it calls addDependent method of every parameter that it depends on.
The method adds the current parameter to dependency list of the
called parameter. Second, during the simulation the dependency list
will be traversed on every iteration in order to calculate new 
partial likelihood for the model. Clearly, it is not allowed to have
duplicate of some parameter in the dependency list.

@author Karri-Pekka Laakso
*/

class parameter : public quantity {
protected:
    /** Current value of this parameter */
    double currentValue;
    
    /** Previous value of this parameter */
    double prevValue;

    /** Dependents of this parameter */
    vector<parameter*> *dependents;

    /**
     	This attribute has the value true, if this parameter
     	has a fixed value read from data file.
    */
    bool data;

    /** 
    	True, if this parameter has a initialization value.
    */
    bool inited;

    /** 
    	Global vector of all SIMULATABLE parameters. That
    	is, if the attribute data has
    	a value true, then this parameter is not 
    	included in this list.
    */
    static vector<parameter*> parameters;

    /**
	Assigns a given value to this parameter.
       	@param value New value for this parameter
      */
    inline void setValue(double value) { currentValue = value; }

    /**
       Proposes a new value for this parameter. The implementation is
       defined in model dependent subclasses.
       @param oldValue The previous value of this parameter
       @return Proposed value for this parameter
      */
    virtual double propose(double oldValue) const = 0;
    
    /**
       	Gives the proposal probability from value to newValue.
	The implementation is defined in model dependent subclasses.	
       	@param value Old value of this parameter
       	@param newValue New proposed value for this parameter
       	@return proposal probability
    */                    
    virtual logprob proposalDensity(double value, double newValue) const = 0;

public:
    /**
       Constructor. Besides creating this parameter, adds it to the
       vector of all simulatable parameters.
       @param n Name of this parameter
      */
    parameter(const string &n);

    /**
       Initializes the value of this parameter
       @param value Initialization value for this parameter
      */
    inline void initialize(double value) { if (!data) { currentValue = value; inited = true; }}

    /**
     	Calls method init, if this parameter has not been inited or
     	read a fixed value from data. Sets the value of attribute inited
     	as true.
      */
    inline void initialize() { if (!data && !inited) init(); inited = true; }

    /**
       Sets the initial value of this parameter according to a proposal
       distribution.
      */
    virtual void init() = 0;

    /**
       	Returns an iterator to the dependency vector
	@return iterator to dependency vector
      */
    virtual vector<parameter*>::const_iterator giveBeginIterator() const;

    /**
       	Returns an iterator to the end of the dependency vector
	@return iterator to end of dependency vector
      */
    virtual vector<parameter*>::const_iterator giveEndIterator() const;

    /**
       Adds a given parameter to the list of dependents of
       this parameter.
       @param *p Parameter that is dependent of this variable
      */
    virtual void addDependent(parameter *p);

    /**
    	Returns the current value of this parameter.
	@return Current value of this parameter
      */
    inline double value() const { return currentValue; }

    /**
	Assigns this parameter a data value or an initial value, 
	if the string "name = value" is found from the input.   
       	@param &sc Scanner 
       	@param *n Name of the variable
       	@param init True if the value is initialization value, false if data.
       	@return True if the sequence was found, otherwise false.
      */
    virtual bool askValue(scanner &sc, const char *n = 0, bool init = false);

    /**
       Gives the probability density of the current value of this
       parameter. The implementation is defined in model
       dependent subclasses.
       @return Density of the current value.
      */
    virtual logprob density() const = 0;

    /**
       	Metropolis-Hastings algorithm. It is implemented as follows.
       	Let \TEX{$X \subset {\cal R}^{n}$} be the state space of
       	all parameters and let us assume, that we are currently in the point
       	\TEX{${\bf x} \in X$}. We denote the index of this
       	parameter in the state space with \TEX{$i$}.
       	First, we draw a new proposal value for this parameter 
       	\TEX{$\hat{x_{i}}$}
       	according to proposal distribution          
       	\TEX{$ propose( \cdot | x_{i})$} (see method propose()).
       	We denote a new proposal vector as
       	\TEX{${\bf \hat{x}} = (x_{1},\ldots,\hat{x_{i}},\ldots,x_{n}) \in X$}.
       	Let \TEX{$dep(i)$} be the set of variables (parameters) 
       	that are dependent of this variable \TEX{$i$} 
       	(see attribute dependents). Second, we calculate probabilites 
       	\TEX{$f(x_{i}, dep(i))$}, \TEX{$f(\hat{x_{i}}, dep(i))$} 
       	and proposal probabilities \TEX{$ pp({\bf x, \hat{x}})$} and
       	\TEX{$pp({\bf \hat{x}, x})$} (see method proposalDensity(a,b)).
       	Now the point \TEX{${\bf \hat{x}}$} is accepted
       	with probability
	\TEX{$ min(1, \frac{f(\hat{x_{i}}, dep(i)) pp({\bf \hat{x}, x})}{f(x_{i}, dep(i)) pp({\bf x, \hat{x}})})$}.
 	Recall that 
	\TEX{$f({\bf x}) = f(x_{i}, dep(i)) f(dep(i)^{c}) \forall {\bf x} \in X$},
	where 
\TEX{$dep(i)^{c} = (x_{j} \in X | j \neq i, x_{j} \notin dep(i))$}
       	@return True if everything is ok, false if -BASS_INFINITY 
       	is achieved in the calculation of partial likelihoods.
      */
    virtual bool MH();  // made virtual for cluster experiment/HT

    /**	Assigns this parameter the given value and sets it as
	data.  
    	@param The new value for this parameter. */
    void setData(double value);

    /** Queries whether this parameter is set as data. */
    bool isData(void) const { return data; }

    /** Indicates if this parameter has been assigned an initial
	value. 
	@return True if this parameter has been initialized. */
    bool isInited() const { return inited; }

    /**	Returns an iterator to the vector of all the simulatable
	parameters. 
	@return The iterator */
    static vector<parameter*>::const_iterator begin() {
	return parameters.begin();
    }

    /**	Returns an iterator to the end of the vector of all the simulatable
	parameters. 
	@return The iterator */
    static vector<parameter*>::const_iterator end() {
	return parameters.end();
    }

    /** The destructor. */
    virtual ~parameter();

    /** 
    	Assigns a previous value to be
    	currentValue and proposes a new value for currentValue
    */
    void proposeAndSetValue() {
	prevValue = currentValue;
	currentValue = propose(prevValue);
    }
    
    /// Restores previous value for this parameter
    inline void restoreOldValue() { currentValue = prevValue; }
    /**
	Probability of proposing the previous value from the current
	value.
	@param currentValue current value of this parameter
	@param prevValue previous value of this parameter
	@return transition probability
    */
    inline logprob proposalDensityBackward() const { 
	return proposalDensity(currentValue, prevValue); 
    }

    /**
	Probability of proposing the current value from the
	previous value.
       	@param prevValue previous value of this parameter
       	@param currentValue Current value of this parameter
       	@return transition probability

      */
    inline logprob proposalDensityForward() const { 
	return proposalDensity(prevValue, currentValue); 
    }
};


class cluster : public parameter {
public:

    cluster(string n) : parameter(n) {}

    virtual vector<parameter*>::const_iterator begin() = 0;
    virtual vector<parameter*>::const_iterator end() = 0;
    
    virtual void printValue(FILE *file);

    virtual bool MH();

    virtual double propose() { ASSERT(false); return 0; }

    virtual logprob proposalDensity(double, double) { ASSERT(false); return 0; }

    virtual logprob density() { ASSERT(false); return 0; }

    virtual ~cluster() {}
};


#endif // PARAMETER_H

// Local Variables:    
// mode: c++	       
// c-basic-offset: 4   
// make-backup-files: t
// case-replace: nil   
// End:		       

