// main.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: main.cc,v 1.39 1999/10/26 10:19:41 jks Exp $
//

#include "main.h"
#include "common.h"
#include "scope.h"
#include "dirent.h"

#include <cstdarg>
#include <ctime>
#include <cstring>
#include <cstdlib>
#include <cerrno>
using namespace std;

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

compiler c = compiler(0, 0);
scope *s = 0;
environment env;

extern int cgparse(void);

extern FILE *cgout, *cgin;

/** struct genFileInfo contains all information on a file generated
    by Bassist */
struct genFileInfo {
    /// The name of the source (template) file
    string sourceFile;
    /// The name of the destination file
    string destinationFile;
};

/** genFiles is a table of all files generated by Bassist. */
const static struct genFileInfo
genFiles[] = {
    { "gen.h",               "gen.h" },
    { "genComposites.h",     "genComposites.h" },
    { "genComposites.cc",    "genComposites.cc" },
    { "genDistributions.h",  "genDistributions.h" },
    { "genDistributions.cc", "genDistributions.cc" },
    { "genFunctions.h",      "genFunctions.h" },
    { "genFunctions.cc",     "genFunctions.cc" },
    { "genGlobals.h",        "genGlobals.h" },
    { "genGlobals.cc",       "genGlobals.cc" },
    { "genParameters.h",     "genParameters.h" },
    { "genParameters.cc",    "genParameters.cc" },
    //{ "genPwcFunctions.h",   "genPwcFunctions.h" },
    //{ "genPwcFunctions.cc",  "genPwcFunctions.cc" },
    { "genRecords.h",        "genRecords.h" },
    { "genRecords.cc",       "genRecords.cc" },
    { "genMain.cc",          "genMain.cc" },
    { "genmakefile",         "Makefile" },
};

//---------------------------------------------------------------------
// compiler::compiler
//  The constructor of the main thing.
//
// --------------------------------------------------------------------

compiler::compiler(int c, char **v)
    : argc(c), argv(v), tmpdir(0), modelN(""),
      modelDef(""), userFile(""),  makeCopy(false), 
      userDir(""), makeCopyForce(false), printDef(false), parseOnly(false), 
      compileOnly(false), condorCompilation(false), verbose(false), 
      parse_errors(false), 
      variables(0), records(0), references(0), functions(0),
      localRecord(0), proposalTargetVariable(0), composites(0), constants(0), 
      parameters(0), file(stdout), datafile(""), infilename(""), prog(""),
      usercflags("")
{
    variables = new vector<variable*>;
    records = new vector<record*>;
    references = new vector<variable*>;
    functions = new vector<function*>;
    composites = new vector<variable*>;
    constants = new vector<variable*>;
    parameters = new vector<variable*>;
}


//---------------------------------------------------------------------
// compiler::printUsage
//  The usage message output if the command line options aren't right.
// 
// --------------------------------------------------------------------

void
compiler::printUsage()
{
    fputs("Usage: bassist [-d DEBUG] [-hvpEcD] [-o file] [-s[dir] [-f]] [-i file]\n"
	  "               [-I<dir>] [-L<dir>] [-l<lib>] bassist_model_file ...\n"
	  "\n"
	  "General options\n"
	  " -h   help (this message)\n"
	  " -v   verbose output\n"
	  "\n"
	  "Model processing options\n"
	  " -p   print model definitions from parse tree\n"
	  " -E   validate model syntax only\n"
	  " -c   compile into C++ only (see also -s)\n"
	  "\n"
	  "Code generation options\n"
	  " -o <file> 	name simulator as \"file\"\n"
	  " -s[dir]   	put simulator source files in directory \"dir\"\n"
	  " -f        	(used with -s) suppress confirmation questions about\n"
	  "           	overwriting files in \"dir\"\n"
	  " -i <file> 	include file \"file\" in the simulator code\n"
	  " -I <dir>  	include directory \"dir\" in header search path\n"
	  " -L <dir>  	include directory \"dir\" in library search path\n"
	  " -l <lib>  	include library \"lib\" in simulator\n"
	  " -C <flags>  include \"flags\" in compilation command\n"
	  " -D          compile for Condor\n"
	  "\n",
	  stderr);
}


//---------------------------------------------------------------------
// compiler::checkEnvironment
//---------------------------------------------------------------------

void
compiler::checkEnvironment()
{
    string myVersion(BASSIST_VERSION);
    string versionFileName(env.bassist_template_dir());
    versionFileName += "/VERSION";
    FILE *versionFile = fopen(versionFileName.c_str(), "r");
    bool opened = false;	// has the file been opened?
    bool success = true;	// does everything look OK so far?

    if (!versionFile) {
	outputDiagnostic(true, "Warning: cannot open file %s: ",
			 versionFileName.c_str());
	success = false;
    } else {
	opened = true;
    }

    char buf[80];

    if (opened)
    {
	char *result = fgets(buf, 80, versionFile);
	if (!result) {
	    outputDiagnostic(true, "Warning: cannot read from file %s: ",
			     versionFileName.c_str());
	    success = false;
	}
    }

    if (opened)
    {
	int result = fclose(versionFile);
	if (result) {
	    outputDiagnostic(true, "Warning: cannot close file %s: ",
			     versionFileName.c_str());
	    success = false;
	}
    }

    if (!success) {
	outputDiagnostic("  The files in that directory may be from a version "
			 "prior to 0.8.6.");
	outputDiagnostic("  You are running version %s.", myVersion.c_str());
    } else {
	string theirVersion = buf;
	if (theirVersion[theirVersion.length()-1] == '\n')
	    theirVersion.erase(theirVersion.length()-1, theirVersion.length());

	if (myVersion.compare(theirVersion) != 0) {
	    outputDiagnostic("Warning: version mismatch!");
	    outputDiagnostic("  You are running version %s.", myVersion.c_str());
	    outputDiagnostic("  The files in directory %s", 
			     env.bassist_template_dir().c_str());
	    outputDiagnostic("  are meant to be used with version %s.", 
			     theirVersion.c_str());
	    success = false;
	}
    }

    if (!success) {
	outputDiagnostic("  Check the following environment variables:");
	outputDiagnostic("  BASSIST_LIBRARY BASSIST_INCLUDE_DIR BASSIST_TEMPLATE_DIR");
    }
}

//---------------------------------------------------------------------
// compiler::init
//  Read the command line options. Return true if parsing succeeded.
//
// --------------------------------------------------------------------

bool
compiler::init()
{
    bool error = false;
    int opt;

    while ((opt = getopt(argc, argv, "d:hvpDEcfo:s::i:I:L:l:C:")) != -1) {
	switch (opt) {
	case 'd':  
	    debugInit(optarg);
	    break;

	case 'h':
	    printUsage();
	    error = true;	// flag an error so we exit
	    break;

	case 'p':
	    printDef = true;
	    break;
	    
	case 'v':
	    verbose = true;
	    break;
	    
	case 'E':
	    parseOnly = true;
	    break;

	case 'c':
	    compileOnly = true;
	    break;
	    
	case 'o':
	    prog = optarg;
	    break;
	    
	case 's':
	    makeCopy = true;
	    if (optarg != 0)
		userDir = optarg;
	    break;

	case 'f':
	    makeCopyForce = true;
	    break;
	    
	case 'i':
	    userFile = optarg;
	    break;
	    
	case 'I':
	case 'L':
	    {
		string path = optarg;
		canonicalizePath(path);
		switch(opt) {
		case 'I': (usercflags += " -I") += path; break;
		case 'L': (usercflags += " -L") += path; break;
		}
	    }
	    break;
	
	case 'l':
	    usercflags += " -l";
	    usercflags += optarg;
	    break;
	    
	case 'C':
	    usercflags += " ";
	    usercflags += optarg;
	    break;

	case 'D':
	    condorCompilation = true;
	    break;
	    
	default:
	    printUsage();
	    error = true;
	}
    }

    if (makeCopyForce && !makeCopy) {
	outputDiagnostic("Warning: option -f is not useful without option -s");
    }

    return !error;		// return true if successful
}


//---------------------------------------------------------------------
// compiler::setModelName
//  
//
// 
// --------------------------------------------------------------------

void
compiler::setModelName(string s)
{
    modelN = s;

    if (prog.length() == 0) {
	prog = modelN + ".sim";
    }
    if (userDir.length() == 0) {
	userDir = modelN + ".src";
    }
}


//---------------------------------------------------------------------
// compiler::checkModelName
//  
//
// 
// --------------------------------------------------------------------

void
compiler::checkModelName(string s)
{
    if (modelN != s)
	outputDiagnostic("Warning: model name is \"%s\", implementation refers to \"%s\"",
			 modelN.c_str(), s.c_str());
}


//---------------------------------------------------------------------
// compiler::saveModelFile
//  
//
// 
// --------------------------------------------------------------------

bool
compiler::saveModelFile(string fileName)
{
    FILE *file = fopen(fileName.c_str(), "r");
    if (file == 0) {
	outputDiagnostic(true, "Error opening file \"%s\" for reading: ", 
			 fileName.c_str());
	return false;
    }

    char tmpBuffer[2001];
    
    do {
	int i = fread(tmpBuffer, sizeof(char), 2000, file);
	if (ferror(file)) {
	    outputDiagnostic(true, "Error reading file \"%s\"",
			     fileName.c_str());
	    return false;
	}

	tmpBuffer[i] = '\0';
	modelDef += tmpBuffer;
    } while (!feof(file));

    if (fclose(file) != 0) {
	outputDiagnostic(true, "Error closing file \"%s\": ", fileName.c_str());
	return false;
    }

    return true;
}


//---------------------------------------------------------------------
// compiler::quoteModelDef
//  Quote away \ and " from model definition
//
// 
// --------------------------------------------------------------------

void
compiler::quoteModelDef()
{
    size_t idx = 0, spot;

    while((spot = modelDef.find_first_of("\\\"", idx)) != modelDef.npos) {
	modelDef.replace(spot, 0, "\\");
	idx = spot+2;
    }
}


//---------------------------------------------------------------------
// compiler::parseInput
//  Parse the input files in order.
//
// 
// --------------------------------------------------------------------

bool
compiler::parseInput()
{    
    if (argc == optind) {
	outputDiagnostic("No model description files given");
 	printUsage();
	return false;
    }

    outputInformation("Parsing input files");

    parse_files(&argv[optind]);
    if (parse_errors)
	return false;
    // Note: In an earlier version, parse errors only caused warnings.
    // I think aborting is the right thing to do here.

    for (int x = optind; argv[x] != 0; x++) {
	saveModelFile(argv[x]);
    }

    for_each(variables->begin(), variables->end(),
	     mem_fun(&variable::findReferences));
    if (parse_errors)
	return false;
    for_each(references->begin(), references->end(),
	     mem_fun(&variable::findReferences));
    if (parse_errors)
	return false;    
    for_each(records->begin(), records->end(),
	     mem_fun(&record::findReferences));
    if (parse_errors)
	return false;

    quoteModelDef();

    return true;
}


//---------------------------------------------------------------------
// compiler::printTree
//  The code generation function
//
// 
// --------------------------------------------------------------------

void
compiler::printTree() {
    printf("\nVariables\n=========\n");

    for (vector<variable*>::iterator
	     i = variables->begin(), end = variables->end();
	 i != end;
	 i++)
	{
	    (*i)->printCode();
	    printf(";\n");
	}

    printf("\nRecords\n=======\n");

    for (vector<record*>::iterator
	     i = records->begin(), end = records->end();
	 i != end;
	 i++)
	{
	    (*i)->printCode();
	    printf(";\n");
	}

    printf("\nFunctions\n=========\n");

    for (vector<function*>::iterator
	     i = functions->begin(), end = functions->end();
	 i != end;
	 i++)
	{
	    (*i)->printCode();
	    printf(";\n");
	}
}


//---------------------------------------------------------------------
// compiler::collectVariables
//  Collect the variables by type to lists.
//
// These lists are needed in the code generation phase.
// --------------------------------------------------------------------

void
compiler::collectVariables()
{
    for(vector<variable*>::iterator
	    i = variables->begin(), end = variables->end();
	i != end;
	i++)
	
	(*i)->collectTypes(composites, constants, parameters);

    for(vector<record*>::iterator
	    i = records->begin(), end = records->end();
	i != end;
	i++)
	
	(*i)->collectTypes(composites, constants, parameters);
}


//---------------------------------------------------------------------
// compiler::generateCode
//  The main loop in code generation.
//
// The files listed in genFiles are gone through one by one, and the
// blocks which begin with genBegin and end with genEnd are replaced
// with the generated code.
// --------------------------------------------------------------------

bool
compiler::generateCode()
{
    FILE *infile, *outfile;
    string filename = "", resFile = "";
    int files = int(sizeof(genFiles)/sizeof(genFiles[0]));

    // Create the 0-scope
    s = new scope();

    // should use mkdtemp if available
    tmpdir = tmpnam(0);
    if (tmpdir == 0) {
	outputDiagnostic(true, "Error naming temporary directory: ");
	return false;
    }
    
    if (mkdir(tmpdir, S_IRUSR | S_IWUSR | S_IXUSR) != 0) {
	outputDiagnostic(true, "Error making temporary directory \"%s\": ",
			 tmpdir);
	return false;
    }
	
    for (int i=0; i < files; ++i) {
	resFile = tmpdir;
	resFile += '/';
	resFile += genFiles[i].destinationFile;

	outfile = fopen(resFile.c_str(), "w");
	if (outfile == NULL) {
	    outputDiagnostic(true, "Error opening file \"%s\" for writing: ",
			     resFile.c_str());
	    return false;
	}
	cgout = file = outfile;
	
	filename = env.bassist_template_dir();
	filename += '/';
	filename += genFiles[i].sourceFile;
	infile = fopen(filename.c_str(), "r");
	if (infile == NULL) {
	    outputDiagnostic(true, "Error opening file \"%s\" for reading: ",
			     filename.c_str());
	    return false;
	}
	cgin = infile;
	infilename = filename;
	
	cgparse();
	fclose(outfile);
	fclose(infile);
    }

    return true;
}


//---------------------------------------------------------------------
// compiler::copyUserFile
//  Copies the user defined file into the compilation directory
//
//
//---------------------------------------------------------------------

bool
compiler::copyUserFile()
{
    if (userFile.length() > 0) {
	string outputFile = tmpdir;
	outputFile += "/genFunctions.cc";

	return copyFile(userFile.c_str(), outputFile.c_str(), true);
    } else {
	return true;
    }
}

//---------------------------------------------------------------------
// compiler::compileSim
//  Compiles the generated model
//
// 
//---------------------------------------------------------------------

bool
compiler::compileSim()
{
    // "cd $tmpdir && make"
    // FIXME: signal handling
    // FIXME: parameterize on "make"

    pid_t cpid;

    cpid = fork();
    if (cpid < 0) {
	// error
	outputDiagnostic(true, "Cannot fork: ");
	return false;
    } else if (cpid == 0) {
	// child

	if (chdir(tmpdir) != 0) {
	    outputDiagnostic(true, "Cannot chdir to \"%s\": ", tmpdir);
	    exit(-1);
	}

	if (!condorCompilation)
	    execl("/bin/sh", "/bin/sh", "-c", "make all", 0);
	else
	    execl("/bin/sh", "/bin/sh", "-c", "make all LINKER=condor_compile\\ egcs++", 0);
	
	// should never return

	outputDiagnostic(true, "Cannot run make: ");
	exit(-1);
    } else {
	// parent
	int status;
	waitpid(cpid, &status, 0); 

	if (WIFEXITED(status)) {
	    // child exited
	    if (WEXITSTATUS(status) != 0) {
		outputDiagnostic("make failed");
		return false;
	    }
	    return true;
	}
	// ordinarily, control doesn't pass this point

	if (WIFSIGNALED(status)) {
	    // child process was signaled
	    outputDiagnostic("make process received signal %d", WTERMSIG(status));
	} else {
	    // twilight zone
	    outputDiagnostic("waitpid terminated with status %d", status);
	}
	return false;
    }
}

//---------------------------------------------------------------------
// compiler::copyBinary
//  Copies the compiled file to the current directory
//
// 
//---------------------------------------------------------------------

bool
compiler::copyBinary()
{
    string s = tmpdir;
    s += '/';
    s += prog;

    return copyFile(s.c_str(), prog.c_str(), true);
}


//---------------------------------------------------------------------
// compiler::copyFiles
//  Copies files from the temporary directory to the user specified
//  source directory, if necessary. Always copies the resulting binary
//  to the current directory.
//---------------------------------------------------------------------

bool
compiler::copyFiles()
{
    bool succ;

    succ = makeDirectory(userDir.c_str());
    if (!succ)
	return false;
    
    DIR *dir;
    
    dir = opendir(tmpdir);
    if (dir == 0)
	succ = false;

    while(succ && dir) {
	struct dirent *entry;
	errno = 0;
	if (0 != (entry = readdir(dir))) {
	    if (entry->d_name[0] != '\0' // Just being defensive
		&& entry->d_name[0] != '.') // Skip dot-files
		{
		    string sourceName = tmpdir;
		    sourceName += '/';
		    sourceName += entry->d_name;

		    string destName = userDir;
		    destName += '/';
		    destName += entry->d_name;
		    
		    copyFile(sourceName.c_str(), destName.c_str(), makeCopyForce);
		}
	} else {
	    // readdir() returned 0: either end of directory or error
	    if (errno != 0) {
		outputDiagnostic(true, "Error reading directory \"%s\": ", tmpdir); 
		succ = false;
	    }
	    closedir(dir);
	    dir = 0;
	}
    }

    return succ;
}

//---------------------------------------------------------------------
// compiler::parse_error
//  Print and log parse errors.
//
// --------------------------------------------------------------------

void
compiler::parse_error(char *format, ...)
{
    printf("%s:%d: ", datafile.c_str(), bassist_lineno());
    va_list ap;
    va_start(ap, format);
    vfprintf(stdout, format, ap);
    va_end(ap);
    printf("\n");
    fflush(stdout);

    parse_errors = true;
}


//---------------------------------------------------------------------
// compiler::reference_error
//  Print and log errors in model references
// 
// --------------------------------------------------------------------

void
compiler::reference_error(const char *dFile, int line, const char *format, ...)
{
    printf("%s:%d: ", dFile, line);
    va_list ap;
    va_start(ap, format);
    vfprintf(stdout, format, ap);
    va_end(ap);
    printf("\n");
    fflush(stdout);

    parse_errors = true;
}


//---------------------------------------------------------------------
// compiler::run
//  The main routine.
//
// --------------------------------------------------------------------

int
compiler::run()
{
    // To abort the run, `goto ABORT'; to stop it successfully, `goto DONE'.

    // If `succ' is true, all is fine.
    // If not, the compilation will be aborted.
    bool succ;

    checkEnvironment();

    succ = init();
    if (!succ) return 1;

    succ = parseInput();
    if (!succ) goto ABORT;
    
    if (printDef)
	printTree();		// never fails

    if (parseOnly) goto DONE;

    outputInformation("Compiling model to C++ files");
    
    collectVariables();		// never fails
    succ = generateCode();     if (!succ) goto ABORT;
    succ = copyUserFile();     if (!succ) goto ABORT;

    if (!compileOnly) {
	outputInformation("Compiling C++ files to simulator program");
	succ  = compileSim();	// no abort check on `succ' yet
	if (succ)
	    succ &= copyBinary();
    }

    if (makeCopy) {
	outputInformation("Copying files to \"%s\"", userDir.c_str());
	succ &= copyFiles();	// note the &=
    }

    // complain if either compilation or copying failed
    if (!succ) goto ABORT;
    // otherwise fall through to DONE
    
DONE:
    // Successful termination
    {
	cleanUp();
	outputInformation("Done");
	// time_t t = time(0);
	// printf("bassist: done at %s\n", asctime(localtime(&t)));
	return 0;    
    }

ABORT:
    // Unsuccessful termination
    {
	cleanUp();
	outputInformation("Terminated unsuccessfully");
	// time_t t = time(0);
	// printf("bassist: FAILED at %s\n", asctime(localtime(&t)));
	return 1;
    }
}

//---------------------------------------------------------------------
// compiler::doInheritance
//---------------------------------------------------------------------

void
compiler::doInheritance() const
{
    //fprintf(stderr, "***** INHERITANCE *****\n");
    for_each(records->begin(), records->end(),
	     mem_fun(&record::inheritFields));
}

//---------------------------------------------------------------------
// compiler::cleanUp
//  Cleans up the possible temp files
//
// 
// --------------------------------------------------------------------

void
compiler::cleanUp()
{
    if (tmpdir != 0) {
	string cmd = "rm -rf ";
	cmd += tmpdir;
	system(cmd.c_str());
    }
}


//---------------------------------------------------------------------
// compiler::~compiler
//  
//
// 
// --------------------------------------------------------------------

compiler::~compiler()
{
    cleanUp();
}




//----------------------------------------
// Functions
//----------------------------------------


//---------------------------------------------------------------------
// copyFile
//  Copies a file. Returns true iff successful.
//---------------------------------------------------------------------

bool
copyFile(const char *source, const char *destination, bool overwrite)
{
    FILE *in, *out;
    bool ok = true;

    if (0 == (in = fopen(source, "r"))) {
	outputDiagnostic(true, "Error opening file \"%s\" for reading: ", source);
	ok = false;
    }
    
    if (ok && !overwrite) {
	struct stat statbuf;
	int r;

	r = stat(destination, &statbuf);
	if (r == 0) {
	    outputDiagnostic("File \"%s\" exists, will not overwrite", destination);
	    fclose(in);
	    ok = false;
	} else if (errno == ENOENT) {
	    // Good: it seems the file doesn't exist.
	} else {
	    outputDiagnostic(true, "Error checking (by stat) file \"%s\": ", destination);
	    fclose(in);
	    ok = false;
	}
    }

    if (ok && 0 == (out = fopen(destination, "w"))) {
	outputDiagnostic(true, "Error opening file \"%s\" for writing: ", destination);
	ok = false;
    }

    if (ok) {
	char buffer[2001];
	int read, written;

	do {
	    read = fread(buffer, sizeof(char), 2000, in);
	    if (read > 0) {
		written = fwrite(buffer, sizeof(char), read, out);
		if (written < read) {
		    outputDiagnostic(true, "Error writing file \"%s\": ", destination);
		    ok = false;
		}
	    } else if (ferror(in)) {
		outputDiagnostic(true, "Error reading file \"%s\": ", source);
		ok = false;
	    }
	} while (ok && read == 2000);

	if (ok) {
	    if (fclose(in) != 0) {
		outputDiagnostic(true, "Error closing file \"%s\": ", source);
		ok = false;
	    }
	    if (fclose(out) != 0) {
		outputDiagnostic(true, "Error closing file \"%s\": ", destination);
		ok = false;
	    }
	}

	if (ok) {
	    struct stat statbuf;

	    if (stat(source, &statbuf) != 0) {
		outputDiagnostic(true, "Error checking (by stat) file \"%s\": ", source);
		ok = false;
	    }

	    if (ok && chmod(destination, 
			    statbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO))) {
		outputDiagnostic(true, "Error setting mode of file \"%s\" to %#03o: ",
				 destination, 
				 statbuf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO));
		ok = false;
	    }
	}
    }

    return ok;
}

//---------------------------------------------------------------------
// makeDirectory
//  If the named directory doesn't exist, makes it. Returns false in error
//  cases; it is not an error if the directory exists.
//---------------------------------------------------------------------

bool
makeDirectory(const char *dirname)
{
    struct stat statbuf;
    int r;
    bool ok = true;

    r = stat(dirname, &statbuf);
    if (r == 0) {
	if ( ! S_ISDIR(statbuf.st_mode) ) {
	    outputDiagnostic(true, "\"%s\" is not a directory", dirname);
	    ok = false;
	}
	// Otherwise, it is a directory and we can leave it alone.
    } else {
	if (errno == ENOENT) {
	    r = mkdir(dirname, 0777);
	    if (r != 0) {
		outputDiagnostic(true, "Cannot make directory \"%s\": ", dirname);
		ok = false;
	    }
	    // Otherwise, we have successfully made the new directory.
	} else {
	    outputDiagnostic(true, "Error while checking (by stat) file \"%s\": ", dirname);
	    ok = false;
	}
    }

    return ok;
}

//---------------------------------------------------------------------
// canonicalizePath
//  Canonicalizes a path name by resolving symbolic links and removing
//  `..' and `.' directories.
//  The parameter pathname is modified.
//---------------------------------------------------------------------

bool
canonicalizePath(string &pathname)
{
    char buffer[PATH_MAX+1], *result = 0;
    // The manual pages aren't exactly clear on whether PATH_MAX includes
    // the trailing NUL, so let's just play it safe.
    bzero(buffer, PATH_MAX+1);

    result = realpath(pathname.c_str(), buffer);

    if (result == 0) {
	outputDiagnostic(true, "Error resolving path name \"%s\": ", pathname.c_str());
	return false;
    }

    pathname = result;
    return true;
}

//----------------------------------------------------------------------
// outputDiagnostic
//----------------------------------------------------------------------

static void
internalOutputDiagnostic(bool useErrno, char *format, va_list ap)
{
    string message = "bassist: ";
    message += format;
    if (useErrno)
	message += strerror(errno);
    message += '\n';

    vfprintf(stderr, message.c_str(), ap);
}

void
outputDiagnostic(bool useErrno, char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    internalOutputDiagnostic(useErrno, format, ap);
    va_end(ap);
}

void
outputDiagnostic(char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    internalOutputDiagnostic(false, format, ap);
    va_end(ap);
}


//----------------------------------------------------------------------
// outputInformation
//----------------------------------------------------------------------

void
outputInformation(char *format, ...)
{
    string message = "bassist: ";
    message += format;
    message += '\n';

    va_list ap;
    va_start(ap, format);
    vfprintf(stdout, message.c_str(), ap);
    va_end(ap);
}

//---------------------------------------------------------------------
// main
//  The one and only.
//
// 
// --------------------------------------------------------------------

int
main(int argc, char **argv) {
    c = compiler(argc, argv);
    
    return(c.run());
}


// Local Variables:
// mode: c++
// make-backup-files: t
// c-basic-offset: 4
// case-replace: nil
// End:
