// main.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: main.h,v 1.21 1999/08/12 14:20:05 jks Exp $
//

#ifndef MAIN_H
#define MAIN_H

#include <string>
#include <vector>
using namespace std;

#include <stdio.h>
#include "variable.h"
#include "record.h"
#include "function.h"
#include "implementation.h"
#include "environment.h"
#include "lib/compiler-extensions.h"

#ifndef BASSIST_VERSION
#error "Bassist version undefined. Did you run configure and make?"
#endif

// Forward declarations
class scope;
class record;

/** The main class of the program.

    A single instance of the compiler class is the main part of the
    compiler; it guides the compilation. Its attributes include the
    most essential data structures, e.g. the parse tree. */
class compiler {
    friend int yyerror(char *);
    friend void parse_files(char **);

    /// The classic program argument count.
    int argc;
    /// The classic program argument vector.
    char **argv;

    /// The name of the temporary compilation directory.
    char *tmpdir;
    ///	The name of the compiled model, set during the parse phase.
    string modelN;
    /// The complete model specification.
    string modelDef;
    /// The name of the user's function file.
    string userFile;

    /** Indicates that the simulation program source code should be
	copied somewhere. */
    bool makeCopy;
    /** The name of the directory, into which the simulation program code
	should be copied. */
    string userDir;
    /** Whether the source code copying should be forced 
        (cp -f as opposed to cp -i). 
	@see makeCopy */
    bool makeCopyForce;

    /** Set, if the user requested a print of the model specification
	accomplished using the parse tree. */
    bool printDef;
    /** If true, the user only wanted to validate the model syntax;
	stop before code generation. */
    bool parseOnly;
    /** If true, the use only wanted to compile the model into C++; stop
        before further compiling the resulting code. */
    bool compileOnly;
    /** If true, compile for use with the Condor system */
    bool condorCompilation;

    /// Set, if the user requested detailed output from the compiler.
    bool verbose;

    /// Indicator of parse errors in the parse phase.
    bool parse_errors;
    
    /// Prints a short help message and exits.
    void printUsage();
    /** Checks that the environment is sane and warns the user if not. */
    void checkEnvironment();
    /** Reads the command line options and inits the flags accordingly.
      Returns true if successful. */
    bool init();
    /** Reads the model specification into the memory (modelDef). 
        @return true iff successful
    	@see modelDef */
    bool saveModelFile(string fileName);
    
    /** Replaces the difficult characters \ and " with \\ and \" in
    	the model specification.  The character replacement is needed,
    	because the model specification is presented as a character
    	string in the simulator code. Therefore, all the special
    	characters should be escaped. Modifies modelDef.
    	@see modelDef */
    void quoteModelDef();

    /** Implements the whole parsing phase of the compiler.

	There is no check for the legality of the variable references
	during the model parsing. This enables the use of cyclic
	dependecies in the model specification. After the whole model
	is parsed, the references must be checked. This is
	accomplished by iterating through the variable and record
	lists and calling the findReferences function of each list
	member. 

	@return true iff successful

	@see variable::findReferences
	@see record::findReferences
	@see expr::findReferences */
    bool parseInput();

    /** Prints the whole model specification from the parse tree to
	the standard output. This is mainly needed in verifying what
	the parse tree looks like after the parse phase. The print has
	been requested by the user, if printDef has been set to true.

	@see node::printCode
	@see printDef */
    void printTree();
    
    /** Collects all the model variables to the type lists. There are
	three type lists, one for stochastic variables (class
	parameter), one for deterministic variables (class composite)
	and one for constants. The type lists are needed in the code
	generation phase.
	@see variable::collectTypes */
    void collectVariables();
    
    /** The code generation phase of the compiler.
	
	At first, a temporary directory for code compilation is created.
	Then, the code template files listed in genFiles are parsed one
	by one by the code generation parser function cgparse(). The
	parser replaces the code generation commands with code from the
	parse tree with the help of cglex() and and the scope class. 

	@return true iff successful
	@see cgparse
	@see scope */
    bool generateCode();
    
    /** Copies the user defined file into the compilation directory. 
     @return true iff successful */
    bool copyUserFile();

    /** Appends the definitions to the command line to invoke
        make. Make gets some information needed in during the
        compilation of the generated code, for example
        "USERFILE=testfunc.cc". 

	@param cmdLine The command line buffer pointer.
	@param def The definition name.
	@param val The definition value.  */
    void appendCmdLine(string &cmdLine, string def, string val);

    /** Compiles the generated program by invoking make.  At first, all the
	information concerning compilation is gathered into a command
	line, which in turn is used to invoke make. This is actually
	the third phase of the compiler. 
	@return true iff successful */
    bool compileSim();

    /** Copies the generated binary file to the current directory.
      @return true iff successful */
    bool copyBinary();

    /** Copies files from the temporary directory. The binary file resulting
        from the compilation is always copied into the current directory,
	and if the user so requests (-s), all of the files are copied into the
	source directory. */
    bool copyFiles();
    
public:
    /** The constructor.
	@param c The argument count (argv) of the program.
	@param v The argument vector (argv) of the program. */
    compiler(int c, char **v);

    /// The list of the global variable definitions of the model.
    vector<variable*> *variables;
    /// The list of the record definitions of the model.
    vector<record*> *records;
    /// The list of the global record references of the model.
    vector<variable*> *references;
    /// The list of the protypes of the user-defined functions of the model.
    vector<function*> *functions;
    /// The local (current) record during model parsing. 
    record *localRecord;
    /// The target variable of the current proposal distribution
    variable *proposalTargetVariable;
    /** The list of the deterministic variables of the model. This list
	is built between the parse phase and the code generation phase.
	@see collectVariables */
    vector<variable*> *composites;
    /** The list of the constants of the model.  This list is built between
	the parse phase and the code generation phase.
	@see collectVariables  */
    vector<variable*> *constants;
    /** The list of the stochastic variables of the model. This list
	is built between the parse phase and the code generation phase.
	@see collectVariables */
    vector<variable*> *parameters;

    /// The file descriptor of the file into which the code is currently generated.
    FILE *file;
    /// The current model specification file name.
    string datafile;
    /// The current code template file in use.
    string infilename;
    /// The user-defined name of the simulator program.
    string prog;
    /** The user defined C compiler flags, including user-defined paths
        and libraries. */
    string usercflags;
    
    /**
    	The main method of the compiler. Executes the three
    	phases of the compiler: parsing (parse\_input),
    	code generation (generateCode) and the C++ compiling
    	phase (compileSim).
	@see parse_input
	@see generateCode
	@see compileSim */
    int run();
    /// Cleans up the possible temporary files and directories before exit.
    void cleanUp();

    /** Sets the model name.
	@param s The model name.
	@see modelN */
    void setModelName(string s);
    /** Checks that the model name is as specified.
        If not, warns the user.
        Used in the implementation parsing phase, this may aid if the user
	keeps model and implementation specifications in different files. */
    void checkModelName(string s);
  
    /** Gives the version of the compiler.
	@return a pointer to the compiler version string */
    string version() { return BASSIST_VERSION; }
    /** Gives the model name.
	@return a pointer to the model name.
	@see modelN
	@see setModelName */
    string modelName() { return modelN; }
    /** Gives a pointer to the model specification.
	@return a pointer to the model specification */
    string modelDefinition() { return modelDef; }
    
    /** Reports a parse error. The file and the line where the error
	occurred are printed preceding the error message, which is in
	printf() form.
	@param format The error message format for printf().
	@param ... The variable list of arguments for the print
	format. */
    void parse_error(char *format, ...);
    /** Reports a variable reference error. The report is like the one 
	generated by parse\_error, but the file and the line number
	where the error occurred must be specified by the caller. The
	references are checked when the whole model specification has
	been parsed, so the global input file name and the line number
	are not valid anymore.
	@see parse_error
	@param dFile The model specification file, where the reference was.
	@param line The line of reference in the model specification
	file.
	@param format The error message format for printf().
	@param ... The variable list of arguments for the print
	format. */
    void reference_error(const char *dFile, int line, const char *format, ...);

    /** Causes the compiler to call inheritFields() for all records. */
    void doInheritance() const;

    /** The destructor. If the scope to be deleted is not the last
	scope, an error occurs and exit() is called. */
    ~compiler();
};


//----------------------------------------
// Global variables
//----------------------------------------

/** @name Global variables */
//@{ 
/// The global compiler instance.
extern compiler c;
/// The scope matching the current (deepest) FOREACH loop.
extern scope *s;
/// The configuration variables settable using environment variables.
extern environment env;
//@}

//----------------------------------------
// Misc utility functions
//----------------------------------------

/** @name Miscellaneous utility functions */
//@{
/** Copies the file named source into destination. Will only overwrite
    the destination file if overwrite == true. */
bool copyFile(const char * source, const char * destination, bool overwrite=false);

/** If the named directory doesn't exist, makes it. Returns false in error
    cases; it is not an error if the directory exists. */
bool makeDirectory(const char *dirname);

/** Canonicalizes a path name by resolving symbolic links and removing
    `..' and `.' directories.
    @param pathname  The pathname to be modified. 
    */
bool canonicalizePath(string &pathname);

/** Parses Bassist model specification files given in an argv-like
    array of strings. Sets c.parse_errors to true if there were errors. */
void parse_files(char **);

/** Returns the number of the line being parsed in the currently open
    input file. */
int bassist_lineno();

/** Outputs a diagnostic to standard error. If useErrno is true,
    the value of errno is used to output an error message right
    after the diagnostic. The rest of the arguments work like
    printf, but don't append a newline to the format string. */
void outputDiagnostic(bool useErrno, char *format, ...) GCC_FORMAT(printf,2,3);
/** Equivalent to outputDiagnostic(false, format, ...); */
void outputDiagnostic(char *format, ...) GCC_FORMAT(printf,1,2);

/** Outputs an informative message to standard output. */
void outputInformation(char *format, ...) GCC_FORMAT(printf,1,2);


//@}

#endif // MAIN_H

//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// c-basic-offset: 4           //
// End:                        //
//-----------------------------//
