%{
// parse-bassist.y
//
// Created 1998 by Karri-Pekka Laakso
//
// $Modified: Fri Aug 28 08:37:17 1998 kpalaaks$
// $Id: parse-bassist.y,v 1.24 1999/10/26 10:19:41 jks Exp $
//

#include <vector>
#include <cstdio>
using namespace std;

#include "common.h"
#include "main.h"
#include "expr.h"
#include "record.h"
#include "variable.h"
#include "implementation.h"
#include "distribution.h"

#define YYDEBUG 1

extern FILE *yyin;
static char **filenames;
extern "C" int yywrap();

// We want line numbers relative to the beginning of this file,
// not across files: yylineno-baselineno+1
extern int yylineno;
static int baselineno;

int
bassist_lineno()
{
    return yylineno-baselineno+1;
}

int 
yyerror(char *msg)
{
    extern int yylineno;
    extern char *yytext;

    c.parse_errors = true;

    outputDiagnostic("%s:%d: %s -- token %s",
		     c.datafile.c_str(), bassist_lineno(), msg, yytext);
    return 0;
}

variable *
resolveVariable(const string &name)
{
    vector<variable*>::iterator
	i = c.variables->begin(),
	end = c.variables->end();
    for (; i != end; i++)
	if ((*i)->giveName() == name)
	    return (*i);
    c.parse_error("undefined variable: %s", name.c_str());
    return 0;
}

variable *
resolveVariable(const string &rec, const string &field)
{
    record *theRecord = 0;

    vector<record*>::iterator
	i = c.records->begin(),
	end = c.records->end();
    for (; i != end; i++)
	if ((*i)->giveName() == rec) {
	    theRecord = *i;
	    break;
	}
    if (theRecord != 0) {
	vector<variable*>::iterator
	    i = theRecord->allFields()->begin(),
	    end = theRecord->allFields()->end();
	for (; i != end; i++)
	    if ((*i)->giveName() == field)
		return (*i);
    }
    c.parse_error("undefined variable: %s.%s", rec.c_str(), field.c_str());
    return 0;
}

extern int yylex(void);


%}

%start model_start

%union {
    string *s;
    int i;
    double f;
    bool b;
    expr *e;
    vector<expr*> *le;
    vector<variable*> *ln;
    record *r;
    variable *v;
    recordRef *rr;
    distributionRef *dr;
}

%token <s> ID
%token <f> FLOAT
%token MODEL VAR RECORD REF AS PREDICTED DISTRIBUTION
%token IMPLEMENTATION INIT PRIOR PROPOSE UPDATE ORDERED UNORDERED
%token AGGR_SUM AGGR_PRODUCT AGGR_MIN AGGR_MAX AGGR_AVG AGGR_COUNT

%left  ','
%left  '+' '-'
%left  '*' '/' 
%right UMINUS
%left  PAREN

%type <e>  expr variable_ref aggregate_function_call
%type <le> param_list expr_list
%type <ln> record_fields
%type <r>  record_def
%type <v>  record_field variable_def impl_var_ref
%type <b>  predicted
%type <rr> record_ref
%type <dr> distribution_ref
%type <i>  dist_param_list

%%

model_start: model_id model_definition 		{ c.doInheritance();
						  outputDiagnostic("Warning: missing implementation specification"); }
	| model_id model_definition implementation_id  { c.doInheritance(); }
  	  implementation_specification          { }
						
	;

model_id: MODEL ID ';' 				{ c.setModelName(*$2); delete $2; }
	;

model_definition: /* empty */
	| model_definition definition
	;

definition: variable_def ';' 			{ c.variables->push_back($1); }
	| record_def 				{ c.records->push_back($1); }
        | record_ref ';'                        { c.references->push_back($1); }
        | distribution_def ';'                  { }
        | /* empty */ ';'                       { }
	; 

distribution_def: DISTRIBUTION ID '(' dist_param_list ')' '=' ID
        { distributionDefinition::newDistribution(
            new distributionDefinition(*$2, $4, *$7, false));
	  delete $2; delete $7;
	}

/* A dist_param_list is the list of parameters given when defining a new
   distribution. We don't want to know the names of the parameters, only
   how many there are. */
dist_param_list: /* empty */                    { $$ = 0; }
        | ID                                    { $$ = 1; }
        | dist_param_list ',' ID                { $$ = $1 + 1; }
        ;

variable_def:  VAR ID '~' distribution_ref predicted
						{ $$ = new stochasticVar(*$2, $4, $5); delete $2; }
	| VAR ID '=' expr		 	{ $$ = new determVar(*$2, $4); delete $2; }
	| VAR ID			 	{ $$ = new variable(*$2); delete $2; }
	;

distribution_ref: ID '(' param_list ')'         { $$ = new distributionRef(*$1, $3); delete $1; }
	;

record_def:  RECORD ID '{'			{ $$ = new record(*$2, new vector<variable*>);
                                                  delete $2;
						  c.localRecord = $$;
						}
	  record_fields '}'			{ $$ = c.localRecord;
						  c.localRecord = 0;
						}
	| RECORD ID ':' ID '{'			{ $$ = new record(*$2, *$4, new vector<variable*>);
						  delete $2; delete $4;
						  c.localRecord = $$;
						}
	  record_fields '}'			{ $$ = c.localRecord;
						  c.localRecord = 0;
						}
	;

predicted: PREDICTED 				{ $$ = true; }
	| /* empty */				{ $$ = false; }
	;

record_fields: /* empty */			{ $$ = c.localRecord->allFields(); }
	| record_fields record_field ';'	{ $1->push_back($2); }
	;

record_field: variable_def 			{ $$ = $1; }
	| record_ref	 			{ $$ = $1; }
	;

record_ref: REF ID		 		{ $$ = new recordRef(*$2); delete $2; }
	| REF ID AS ID			 	{ $$ = new recordRef(*$2, *$4); delete $2; }
	;

variable_ref: ID '.' ID			{ $$ = new variableRef(*$1, *$3); delete $1; delete $3; }
	| ID				{ $$ = new variableRef(*$1); delete $1;}
	;

expr:	  expr '+' expr			{ $$ = new arithmeticExpr($1, PLUS, $3); }
	| expr '-' expr			{ $$ = new arithmeticExpr($1, MINUS, $3); }
	| expr '*' expr			{ $$ = new arithmeticExpr($1, MULT, $3); }
	| expr '/' expr			{ $$ = new arithmeticExpr($1, DIV, $3); }
	| '-' expr %prec UMINUS		{ $$ = new negationExpr($2); }
	| '(' expr ')' %prec PAREN	{ $$ = new parenExpr($2); }
	| variable_ref			{ $$ = $1; }
	| ID '(' param_list ')'	        { $$ = new functionCall(*$1, $3); delete $1; }
	| aggregate_function_call       { $$ = $1; }
	| FLOAT				{ $$ = new constant($1); }
	;

param_list: expr_list 			{ $$ = $1; }
	| /* empty */                   { $$ = new vector<expr*>; /* Empty paramlist */}
	;

expr_list: expr_list ',' expr 		{ $1->push_back($3); }
	| expr		 		{ $$ = new vector<expr*>; $$->push_back($1); }
	;

aggregate_function_call: 
	  AGGR_COUNT '(' ID ')' 	{ $$ = new aggregateFunctionCall("count", *$3); 
	                                  delete $3; }
	| AGGR_SUM '(' ID '.' ID ')' 	{ $$ = new aggregateFunctionCall("sum", *$3, *$5);
	                                  delete $3; delete $5; }
	| AGGR_PRODUCT '(' ID '.' ID ')' { $$ = new aggregateFunctionCall("product", *$3, *$5);
	                                  delete $3; delete $5; }
	| AGGR_MIN '(' ID '.' ID ')' 	{ $$ = new aggregateFunctionCall("min", *$3, *$5);
	                                  delete $3; delete $5; }
	| AGGR_MAX '(' ID '.' ID ')' 	{ $$ = new aggregateFunctionCall("max", *$3, *$5);
	                                  delete $3; delete $5; }
	| AGGR_AVG '(' ID '.' ID ')' 	{ $$ = new aggregateFunctionCall("avg", *$3, *$5);
	                                  delete $3; delete $5; }
	;

implementation_id: IMPLEMENTATION ID ';'	{ c.checkModelName(*$2); delete $2; }
	;

implementation_specification: impl_clause ';'
	| implementation_specification impl_clause ';'
	;

impl_clause: init_clause
	| proposal_clause
	| update_clause			{ outputDiagnostic("Warning: update clauses unimplemented"); /*FIXME*/}
	;

init_clause: INIT impl_var_ref '~' { c.localRecord = $2->giveRecord(); } distribution_ref 
				   { $2->addInitClause(new randomInitializationClause($5, $2)); 
				     c.localRecord = 0; }
	| INIT impl_var_ref '~' PRIOR   { $2->addInitClause(0); }
	| INIT impl_var_ref '=' { c.localRecord = $2->giveRecord(); }
	  expr { $2->addInitClause(new constantInitializationClause($5, $2)); 
	         c.localRecord = 0; }
	;

proposal_clause: PROPOSE impl_var_ref '~' { c.localRecord = $2->giveRecord(); 
					    c.proposalTargetVariable = $2; }
	distribution_ref            { if ($2 != 0) 
					$2->addProposalClause(new proposalClause($5, $2)); 
				      c.localRecord = 0; 
				      c.proposalTargetVariable = 0; }
	| PROPOSE impl_var_ref '~' PRIOR
					{ if ($2 != 0) $2->addProposalClause(); }
	;

update_clause: UPDATE ORDERED
	| UPDATE UNORDERED		
	;

impl_var_ref: ID '.' ID			{ $$ = resolveVariable(*$1, *$3); delete $1; delete $3; }
	| ID				{ $$ = resolveVariable("", *$1); delete $1; }
	;

%%

int
yywrap()
{
    // Close former input file. 
    // By default, yyin==stdin according to some documentation, 
    // but it seems to be ==0 in our implementation.
    if (yyin != stdin && yyin != 0) { 
	fclose(yyin);
    }
    yyin = 0;

    // Open consecutive files until there are no more files
    // or opening one file succeeds.
    while (*filenames != 0 && yyin == 0) {
	yyin = fopen(*filenames, "r");
	if (yyin == 0) {
	    outputDiagnostic(true, "Cannot open file \"%s\" for reading: ", 
			     *filenames++);
	} else {
	    // Succeeded
	    outputInformation("Reading file \"%s\"", *filenames);
	    c.datafile=*filenames++;
	    baselineno=yylineno;
	    return 0;
	}
    }

    ASSERT(*filenames == 0);

    // No more files
    c.datafile="";
    baselineno=yylineno;
    return 1;
}

void
parse_files(char **files)
{
    //yydebug = 1;
    filenames = files;

    if (yywrap() != 0) {		// Open the first file
	outputDiagnostic("No files to process");
	c.parse_errors = true;
	return;
    }

    yyparse();
}

// Local Variables:
// mode: c++
// c-basic-offset: 4
// case-replace: nil
// make-backup-files: t
// End:
