// record.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: record.cc,v 1.11 1999/08/17 18:41:35 jks Exp $
//

#include "common.h"
#include "record.h"
#include "main.h"

#include <algorithm>
using namespace std;

#include <stdio.h>

extern int yylineno;


//---------------------------------------------------------------------
// record::record
//  The record class ordinary constructor.
//
// 
// --------------------------------------------------------------------

record::record(const string &n, vector<variable*> *f) : name(n), type("record_"),
    base(0), fields(f), recRefs(new vector<variable*>), vars(new vector<variable*>),
    ownVars(0), haveInherited(false)
{
    type += name;

    if (recordByName(n))
	c.parse_error("redefinition of record '%s'", n.c_str());
}

//---------------------------------------------------------------------
// record::record
//  The record class derived record constructor.
//
// 
// --------------------------------------------------------------------

record::record(const string &n, const string &b, vector<variable*> *f) :
    name(n), type("record_"), base(recordByName(b)), fields(f), 
    recRefs(new vector<variable*>), vars(new vector<variable*>), 
    ownVars(0), haveInherited(false)
{
    type += name;

    if (recordByName(n))
	c.parse_error("redefinition of record '%s'", n.c_str());

    if (!base)
	c.parse_error("record '%s' derived from nonexistent record '%s'", 
		      n.c_str(), b.c_str());
}

//---------------------------------------------------------------------
// record::recordByName
//  Find the record called 'name'.
//
// This function is used to avoid redefinition of existing records,
// and to find the base records when defining a derived record.
// --------------------------------------------------------------------

record *
record::recordByName(const string &name)
{
    vector<record*>::iterator
	i = c.records->begin(),
	end = c.records->end();
    for (; i != end; i++)
	if (name.compare((*i)->giveName()) == 0) 
	    return *i;
    
    return 0;
}

//----------------------------------------------------------------------
// record::ownVariables()
//----------------------------------------------------------------------

vector<variable*> *
record::ownVariables()
{
    if (ownVars != 0)		// have cached copy
	return ownVars;
    if (base == 0)		// not a derived record
	return vars;

    ownVars = new vector<variable*>;
    vector<variable*>::iterator 
	baseBegin = base->vars->begin(),
	baseEnd = base->vars->end(),
	j,
	i = vars->begin(),
	end = vars->end();
    for (; i != end; i++) {
	bool matched = false;
	const string &name = (*i)->giveName();
	for (j = baseBegin; j != baseEnd; j++)
	    if (name == (*j)->giveName()) {
		matched = true;
		break;
	    }
	if (!matched)
	    ownVars->push_back(*i);
    }

    return ownVars;
}


//---------------------------------------------------------------------
// record::inheritFields
//  
//
// 
// --------------------------------------------------------------------

void
record::inheritFields()
{
    // Cause the function to be idempotent
    if (haveInherited)
	return;
    haveInherited = true;

    // Inheritance is only sensible on derived records
    if (!base)
	return;

    // Make sure the base has inherited
    base->inheritFields();

    // FIXME: this is O(mn), could be more intelligent
    // For all fields of the base record:
    vector<variable*>::iterator
	i = base->fields->begin(),
	end = base->fields->end();
    for (; i != end; i++) {
	bool foundMatch = false;
	vector<variable*>::iterator
	    j = fields->begin(),
	    jEnd = fields->end();

	// See if we already have a field with that name. If not,
	// create a new variable.
	for (; j != end; j++) {
	    if ((*i)->giveName() == (*j)->giveName()) {
		foundMatch = true;
		break;
	    }
	}
	if (!foundMatch) {
	    fields->push_back((*i)->createInherited(this));
	}
    }
}

//---------------------------------------------------------------------
// record::findReferences
//  
//
// 
// --------------------------------------------------------------------

void
record::findReferences()
{
    for_each(fields->begin(), fields->end(),
	    mem_fun(&variable::findReferences));
}


//---------------------------------------------------------------------
// record::printCode
//  The code generating function.
//
// 
// --------------------------------------------------------------------

void
record::printCode()
{
    if (base)
	fprintf(c.file, "record %s: %s {", name.c_str(), base->name.c_str());
    else
	fprintf(c.file, "record %s {", name.c_str());

    vector<variable*>::iterator
	i = fields->begin(),
	end = fields->end();
    for (; i != end; i++) {
	(*i)->printCode();
	if (i+1 != end)
	    fprintf(c.file, ", ");
    }
    fprintf(c.file, "}");
}


//---------------------------------------------------------------------
// record::collectTypes
//  
//
// 
// --------------------------------------------------------------------

void
record::collectTypes(vector<variable*> *c, vector<variable*> *o,
		     vector<variable*> *s)
{
    vector<variable*>::iterator
	i = fields->begin(),
	end = fields->end();
    for (; i != end; i++) {
	(*i)->collectTypes(c, o, s);
	(*i)->collectRecFields(recRefs, vars);
    }
}


//----------------------------------------
// recordRef - record reference class
//----------------------------------------



//---------------------------------------------------------------------
// recordRef::recordRef
//  The record reference class constructors.
//
// 
// --------------------------------------------------------------------

recordRef::recordRef(const string &r) : 
    variable(r.c_str()), recName(r), dFile(c.datafile), line(yylineno), rec(0)
{
    type = "record_" + recName;
    // The reference check is done afterwards
}


recordRef::recordRef(const string &r, const string &a) : 
    variable(a.c_str()), recName(r), dFile(c.datafile), line(yylineno), rec(0)
{
    type = "record_" + recName;
    // The reference check is done afterwards
}

// recordRef::createInherited
// almost like a copy constructor -- FIXME: should it be?
// uses new to allocate the return value

variable *
recordRef::createInherited(record *r) const
{
    ASSERT(myRecord != 0);	// can only be called for record vars
    ASSERT(c.localRecord == 0); // should only get called after parsing
    c.localRecord = r;
    variable *v = new recordRef(recName, name);
    c.localRecord = 0;

    return v;
}


//---------------------------------------------------------------------
// recordRef::findReferences
//  Finds the references to other variables.
//
// Reference finding is done after the compilation to make
// cross-references possible. 
// --------------------------------------------------------------------

void
recordRef::findReferences()
{
    if (rec == 0) {
	// Search for the record
	vector<record*>::iterator
	    i = c.records->begin(),
	    end = c.records->end();
	for (; i != end; i++)
	    if (recName == (*i)->giveName()) {
		rec = (*i);
		break;
	    }

	// If the record was not found, we define it to prevent further
	// errors. 
	if (rec == 0) {
	    rec = new record(recName.c_str(), new vector<variable*>);
	    c.records->push_back(rec);

	    c.reference_error(dFile.c_str(), line, "record '%s' is undefined", recName.c_str());
	}
    }
}


//---------------------------------------------------------------------
// recordRef::nameSpace
//  
//
// 
// --------------------------------------------------------------------

vector<variable*>* 
recordRef::nameSpace()
{
     if (rec == 0)
	 findReferences();
     
     return rec->allFields(); 
}


//---------------------------------------------------------------------
// recordRef::printCode
//  The code generating function.
//
// 
// --------------------------------------------------------------------

void
recordRef::printCode() const
{
    fprintf(c.file, "ref %s", rec->giveName().c_str());    
    if (rec->giveName().compare(name) != 0)
	fprintf(c.file, " as %s", name.c_str());
}


//---------------------------------------------------------------------
// recordRef::genPointer
//  
//
// 
// --------------------------------------------------------------------

void
recordRef::genPointer()
{
    fprintf(c.file, "record_%s *%s;", rec->giveName().c_str(), codename.c_str());
}


//---------------------------------------------------------------------
// recordRef::genRecName
//  
//
// 
// --------------------------------------------------------------------

void
recordRef::genRecName()
{
    fprintf(c.file, "%s", rec->giveName().c_str());
}


// Local Variables:    
// mode: c++           
// make-backup-files: t
// c-basic-offset: 4   
// case-replace: nil   
// End:                
