// genGlobals.cc
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: genGlobals.cc,v 1.17 1999/08/16 13:13:57 jks Exp $
//

#include "genGlobals.h"

#include "main.h"
#include "common.h"
#include "quantity.h"
#include "parameter.h"
#include "constant.h"
#include "scanner.h"

#include "genFunctions.h"
#include "genParameters.h"
#include "genRecords.h"

map<string, quantity*> *globalConstants; 
map<string, quantity*> *globalParameters;

// Instantiate all the pointers to the global variables
// 
// For example:
// constant *n;       
// ...
// parameter *prob_p;
// ...
// composite *alpha;   <--- composites
// ...

// FOREACH VAR (GLOBAL_VAR)
$VAR.POINTER
// ENDFOR

// FOREACH REF (GLOBAL_REF)
$REF.POINTER
char *$REF.VNAMERef = 0;
// ENDFOR

//---------------------------------------------------------------------
// simulator::printModelName
//  Print the model definitions.
//
// --------------------------------------------------------------------

void
simulator::printModelName(FILE *file)
{
    fprintf(file, "# Bassist @VERSION, model '@MODEL', compiled @DATE\n");
}


//---------------------------------------------------------------------
// simulator::createGlobals
//  Create the global variables of the simulation program.
//
// --------------------------------------------------------------------

void
simulator::createGlobals()
{
    // These maps are used to store the globals
    globalConstants = new map<string, quantity*>;
    globalParameters = new map<string, quantity*>;

// The creation of the instances of global variables. The global
// parameters and constants are added to the maps. 
// For example
//	  n      = new constant(100000, "n");
//	  prob_p = new prob_p_parameter("prob_p");
//	  alpha  = new alpha_determ("alpha");
//
//	  (*globalParameters)[prob_p->identifier()] = prob_p;
//	  (*globalConstants)[alpha->identifier()] = alpha;

// FOREACH VAR (GLOBAL_VAR)
    $VAR.CREATE
    $VAR.ADD_TO_LIST

// ENDFOR
}


//---------------------------------------------------------------------
// simulator::initGlobals
//  Initialization of all globals in the model. 
//
// This must be done before other initializations, because the other
// init values might (= they will) depend on values that are
// calculated here.
//
// Problem: if there are complex dependencies between the constants,
// the initialization might go in wrong order. This is NOT checked
// here in any way.
// --------------------------------------------------------------------

void
simulator::initGlobals()
{
// FOREACH REF (GLOBAL_REF)
    if ($REF.VNAMERef == 0) {
	fprintf(stderr,
	        "Missing reference definition for \"$REF.VNAME\"\n");
	exit(1);
    }
    $REF.NAME = $REF.TYPE::findReference($REF.VNAMERef);
    if ($REF.NAME == 0) {
	fprintf(stderr,
		"Instance \"%s\" for reference \"$REF.VNAME\" not found\n",
	        $REF.VNAMERef);
	exit(1);
    }
    delete $REF.VNAMERef; $REF.VNAMERef = 0;

// ENDFOR

// FOREACH VAR (GLOBAL_VAR)
    $VAR.DEPENDENCIES
// ENDFOR

// FOREACH VAR (GLOBAL_VAR)
    $VAR.NAME->initialize();
// ENDFOR
}


//---------------------------------------------------------------------
// simulator::initRecords
//  The initialization of records in the model.
//
// This is the final task in the simulation initialization. Every
// record variable should report its dependency to the ones it depends
// on to ensure that when the values of those variables are changed,
// the change in this variable's likelihood is taken into account in
// the MH acceptance ratio. Every variable also needs a starting
// value (except composites).
// --------------------------------------------------------------------

void
simulator::initRecords()
{
    // First, check that all records have at least one instance in the data.
// FOREACH REC (RECORD)
    if (record_$REC.NAME::instanceMap == 0 ||
	record_$REC.NAME::instanceMap->empty()) {
	fprintf(stderr, "Warning: No records of type $REC.NAME.\n");
    }
// ENDFOR

    // Next, the links are initialized.
// FOREACH REC (RECORD)
    if (record_$REC.NAME::instanceMap != 0) {
	for (record_$REC.NAME::instanceMapType::iterator p = record_$REC.NAME::instanceMap->begin();
	     p != record_$REC.NAME::instanceMap->end();
	     ++p)
	    {
		p->second->initLinks();
	    }
    }

// ENDFOR

    // Then the record internals are initialized. 
// FOREACH REC (RECORD)
    if (record_$REC.NAME::instanceMap != 0) {
	for (record_$REC.NAME::instanceMapType::iterator p = record_$REC.NAME::instanceMap->begin();
	     p != record_$REC.NAME::instanceMap->end();
	     ++p)
	    {
		p->second->initialize();
	    }
    }

// ENDFOR

    // As this should be the point after which the actual simulation
    // is started, the maps are no longer needed.
    delete globalConstants;  globalConstants  = 0;
    delete globalParameters; globalParameters = 0;
// FOREACH REC (RECORD) 
    if (record_$REC.NAME::instanceMap != 0) {
	delete record_$REC.NAME::instanceMap;
	record_$REC.NAME::instanceMap = 0;
    }
// ENDFOR
    
}


//---------------------------------------------------------------------
// simulator::printModelDef
//  Print the model definition
// 
// --------------------------------------------------------------------

void
simulator::printModelDef(FILE *file)
{
    static const char * const MODELDEF = 
@MODELDEF;
    
    fputs(MODELDEF, file);
}

//---------------------------------------------------------------------
// printDataExample
//  Print an example of the data this simulator takes
// 
// --------------------------------------------------------------------

void
simulator::printDataExample(FILE *file)
{
    fputs("# A data example for model '@MODEL'\n"
	  "# Comment lines begin like these, with a '#'\n"
	  "# Data can be separated to multiple lines, but newlines are not\n"
	  "# allowed in variable names or their decimal values.\n"
	  "\n"
	  "# Every global can be assigned a value:\n"
	  , file);

   double d = 1.0;
// FOREACH VAR (GLOBAL_VAR)
   if (d < 5.0)
       fprintf(file, "$VAR.VNAME = %f ", d++);
// ENDFOR

   fputs("\n\n"
	 "# All globals that are assigned a value become data constants in the\n"
	 "# simulation. If you want to assign an initial value, put an 'i'\n"
	 "# before the value, like\n"
	 "alpha = i1.0\n"
	 "\n"
	 "# In case an initial or data value is not presented, the initial value \n"
	 "# will be taken from the variables prior distribution.\n"
	 "\n"
	 "#Each record data structure consist of the record's type, name,\n"
	 "#references to other records in the same order as in model. References\n"
	 "#are done by other records names. Name for the record instance is not\n"
	 "#compulsory, but it's needed, if the record definition is referres\n"
	 "#to. Data values for the record's other fields the references are\n"
	 "#between the {} in arbitrary order. The system is the sama as with\n"
	 "#globals: only data fields need to be assigned a value, other may take\n"
	 "#initial values, or in case neither exist, the initial value is taken\n"
	 "#from the prior.\n"
	 "\n\n", file);

   int i;
   bool first;

   // These two lines are needed to keep the compiler happy, if there
   // are no records 
   i = 0;
   first = true;
   
// FOREACH REC (RECORD) 
   fputs("$REC.NAME $REC.NAME_ex ", file);
   i = 0;
// FOREACH REF (REC_REFFIELD)
   i++; 
// ENDFOR
   if (i > 0) {
       fputs("[ ", file);
// FOREACH REF (REC_REFFIELD)
       fputs("$REF.RECNAME_ex", file);
       if (--i > 0)
	   fputs(", ", file);
// ENDFOR
       fputs(" ] ", file);
   }
   
   i = 0;
// FOREACH VAR (REC_VARFIELD)
   i++; 
// ENDFOR
   fputs("{ ", file);
   d = 0.0;
   first = true;
// FOREACH VAR (REC_VARFIELD)
   if (strstr("$VAR.TYPE", "composite") == 0) {
       if (!first)
	   fputs(", ", file);
       fprintf(file, "$VAR.VNAME = %f", d++);
       first = false;
   }
// ENDFOR		    
   fputs(" }\n\n", file);
// ENDFOR
}

//---------------------------------------------------------------------
// simulator::readInput
//  Read data or initial values from file
// 
// --------------------------------------------------------------------

void
simulator::readInput(char *file, bool init) {
    DEBUG('s', "# Input file is %s\n", file);

    scanner sc(file);

    while(!sc.eof()) {
// All the read functions for generated records + the global read function

	if (sc.ask(",")
// FOREACH REC (RECORD)
	    || record_$REC.NAME::readInput(sc, init)
// ENDFOR
	    || globalReadInput(sc, init)) {
	  // ok, do nothing
	} else {
	    if (!sc.eof()) {
		fprintf(stderr, "Error in %s:%d: unknown input \"%s\"\n",
		        file, sc.line(), sc.next());
		exit(1);
	    }
	}
    }
    
    DEBUG('s', "# Input loaded\n");
}


// FIXME: The following should probably be done in a nicer, OO, way.

//---------------------------------------------------------------------
// readGlobalReference
//  Initialize a global reference.
//  Used by simulator::globalReadInput only.
// --------------------------------------------------------------------

static bool readGlobalReference(scanner &, char *, char *&) GCC_UNUSED;

static bool
readGlobalReference(scanner &sc, char *id, char *&refName)
{
    char *s = 0;

    sc.get(id);
    sc.get("=");
    s = sc.next();

    if (refName != 0) {
	fprintf(stderr, "%s:%d: warning: redefinition of reference %s\n",
	        sc.fileName(), sc.line(), id);
	delete refName; refName = 0;
    }

    refName = new char[strlen(s)+1];
    strcpy(refName, s);

    return true;
}

//---------------------------------------------------------------------
// simulator::globalReadInput
//  Try to read global variable values from file
// 
// --------------------------------------------------------------------

bool
simulator::globalReadInput(scanner &sc, bool init)
{
    char *s;
    quantity *q;
    
    if (!(s = sc.next(false)))
	return false;

    q = (*globalParameters)[s];
    if (q != 0) 
	return q->askValue(sc, 0, init);  // This should always be true

    q = (*globalConstants)[s];
    if (q != 0) 
	return q->askValue(sc, 0, init);  // This should always be true

// FOREACH REF (GLOBAL_REF)
    if (strcmp(s, "$REF.VNAME") == 0) {
	return readGlobalReference(sc, "$REF.VNAME", $REF.VNAMERef);
    }
// ENDFOR

    return (false);
}


//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// c-basic-offset: 4           //
// make-backup-files: t        //
// End:                        //
//-----------------------------//
