// variable.h
//
// Created 1998 by Karri-Pekka Laakso
//
// $Id: variable.h,v 1.18 1999/10/22 14:03:10 jks Exp $
//

#ifndef VARIABLE_H
#define VARIABLE_H

#include "node.h"
#include "expr.h"
#include "distribution.h"
#include "implementation.h"

// Forward declarations
class record;
class initializationClause;
class proposalClause;

/** A base class for all the variable definitions.

    Every variable definition in the model is of type variable. The
    constants defined in the model are instances of this class, 
    whereas the stochastic variables are instances of class
    stochasticVar and deterministic variables are instances of class
    determVar, both of which are subclasses of this class.  
*/
class variable : public parseTreeNode {
protected:
    /// The name of the variable used in the model specification.
    string name;
    /** The name of the variable used in the generated C++ code. The
	C++ name is the same as the model specification name, but it
	has a 'var\_' in the beginning to avoid name conflicts with
	the rest of the generated simulation program. */  
    string codename;
    /** The class type used in the generated C++ code. The default
	type is 'constant'. */
    string type;
    /** Pointer to the record in which this variable is defined, if
	this variable has been defined in a record structure. */
    record *myRecord;
    /** Set of all aggregate functions used on this variable. */
    set<string> usedAggregates;

public:
    
    /** The constructor.
	When a new variable definition is created, it is checked that
	there doesn't already exist a variable having the same name. If
	one exists, a parse error is reported.
	@param n The name of the variable.
	@see isRedefinition
	@see compiler::parse_error */
    variable(const string &n);

    // /** The destructor -- probably not needed */
    //virtual ~variable();

    /// Does nothing except for stochastic variables.
    virtual void addInitClause(initializationClause *) { };
    /// Does nothing except for stochastic variables.
    virtual void addProposalClause(proposalClause * = 0) { };
    /// Creates an inherited variable in another record, allocates with new
    virtual variable *createInherited(record *r) const;

    /** Gives the name of this variable.
	@see name
	@return a pointer to the name of this variable */
    inline const string &giveName() const { return name; }
    /** Gives the C++ name of this variable.
	@see codename
	@return a pointer to the C++ name of this variable */
    inline const string &giveCodeName() const { return codename; }
    /** Gives the C++ type name of this variable.
	@see type
	@return a pointer to the C++ type name  */
    inline const string &giveType() const { return type; }
    /** Gives a pointer to the record in which this variable has been
	defined.
	@see myRecord
	@return a pointer to the record, or 0 if this is a global
	variable */
    inline record *giveRecord() const { return myRecord; }

    /** Tests if a variable having the same name has been previously
	defined.
	@see variable::variable
	@return true if a variable with the same name exists */
    virtual bool isRedefinition() const;

    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. Default implementation is to do nothing.
	@see compiler::parseInput
	@see variableRef::findReferences */
    inline virtual void findReferences() {}
    
    /** Gives a pointer to the list of the variable definitions inside
        a record. 
        @return a pointer to the list, if this is a record reference,
	otherwise 0
	@see recordRef::nameSpace */
    inline virtual vector<variable*>* nameSpace() { return 0; }

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    virtual void printCode();

    /** Collects the variables of the model into the type lists. The
        variable adds itself to the type list corresponding to its C++
	type. The default implementation assumes that the type is
	"constant".
	@see compiler::collectVariables */
    virtual void collectTypes(vector<variable*> *,
			      vector<variable*> *c,
			      vector<variable*> *) 
    { c->push_back(this); }

    /** Collects the variables defined inside a record into two lists, a
        list of the variable definitions and a list of the record
	references. The default implementation assumes that this is a
	variable definition.
	@see record::collectTypes */
    virtual void collectRecFields(vector<variable*> *,
				  vector<variable*> *v) 
    { v->push_back(this); }

    /** Called typically from aggregateFunctionCall::findReferences to register
      that the aggregate is used on this variable. */
    virtual void registerAggregateUse(const string &function);
    
    /// Generates a C++ pointer. 
    virtual void genPointer();
    /// Generates the code that creates a new instance of this variable. 
    virtual void genCreate();
    /** Generates the code that inserts this variable into a global map. */
    virtual void genAddToList();
    /// Generates the addDependent calls.
    virtual void genDependencies() {}
    /// Generates the initialization function call.
    virtual void genInit() {}
    /** Generates a record attribute to the class definition of this
	variable, if this variable is defined in a record. */
    virtual void genRecConsDef();
    /** Generates a record parameter into the constructor call of this
	variable, if this variable is defined in a record. */ 
    virtual void genRecConsParam();
    /** Generates the initialization of the record attribute in the
	constructor of this variable, if this variable is defined in a
	record. */ 
    virtual void genRecConsInit();
    /** A terrible hack.
      @see aggregateFunctionCall::collectVars */
    virtual void genDependencyHackH();
    /** A terrible hack.
      @see aggregateFunctionCall::collectVars */
    virtual void genDependencyHackC();
    /** */
    virtual void genAggregatePrototypes();
    /** */
    virtual void genAggregateCode();
    /** Is this variable implicit? */
    virtual bool isImplicit() { return false; }
};


/// A class for the stochastic variables of the model.
class stochasticVar : public variable {
    /// The distribution of this variable.
    distributionRef *dist;

    /** Indicates if this parameter should be only predicted, i.e. its
	likelihood should not be taken into account in the
	Metropolis-Hastings probability of the variables it depends
	on. */
    bool predicted;

    /** A counter for implicitly generated composites. The current
	value of this counter becomes a part of the name of the next
	implicit composite. */
    static int genNumber;
    
    /// Possible initialization clause
    initializationClause *initClause;
    /// Possible proposal clause
    proposalClause *propClause;
    
public:
    /** The constructor.
        (The following isn't quite right; some of this has been
	moved to the distributionReference class.)
	At first, the constructor checks that the distribution is
	defined; if it is not, a parse error is reported. The number
	of parameters given to the distribution must also match the
	definition of the distribution. If there are expressions
	that should be represented as implicit composites among the
	parameters of the distribution, the implicit composite
	definitions are created and the expressions are replaced with
	variable references to these composites.
	@see expr::isComposite
	@param n The name of this variable.
	@param d The distribution reference.
	@pred pred Indicates if this variable should be only
	predicted. */
    stochasticVar(const string &n, distributionRef *d, bool pred);
    
    /// Creates an inherited variable in another record, allocates with new
    virtual variable *createInherited(record *r) const;
    /** Adds an initialization clause.
        If i is a null pointer, causes the 
	prior distribution to be used for initialization. */
    void addInitClause(initializationClause *i = 0);
    /** Adds a proposal clause.
        If p is a null pointer, causes the 
	prior distribution to be used for proposals. */
    void addProposalClause(proposalClause *p = 0);
    
    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. The function call is repeated recursively to all of
	the variable references on the parameter list of the distribution.
	@see compiler::parseInput
	@see variableRef::findReferences */
    void findReferences();

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();

    /** Collects the variables of the model into the type lists. The
        variable adds itself to the type list corresponding to its C++
	type. The C++ type of all the stochasticVars is "parameter".
	@see compiler::collectVariables */
    virtual void collectTypes(vector<variable*> *, 
			      vector<variable*> *,
			      vector<variable*> *s) 
    { s->push_back(this); };

    /// Generates the distribution parameters.
    void genFuncParams();
    /// Generates the distribution name.
    void genDistName();
    
    /// Generates the code that creates a new instance of this variable. 
    void genCreate();
    /** Generates the code that inserts this variable into a global map. */
    void genAddToList();
    /// Generates the addDependent calls.
    void genDependencies();
    /// Generates the initialization function call.
    void genInit();
    /// Generates the expression used to initialize this variable.
    void genInitializer();
    /// Generates the expression used to propose a new value.
    void genProposer();
    /// Generates the proposal density call.
    void genProposalDensity();
};



/// A class for deterministic variables of the model.
class determVar : public variable {
    /// The expression that defines the value of this variable.
    expr *rule;
    /** Indicates if this is an implicitly defined deterministic
	variable. */
    bool implicit;
    
public:
    /** The constructor.
	If the expression that defines the value of this variable is a
	constant expression, this variable is a constant type variable
	in the simulator program, otherwise this is a composite.
	@param n The name of this variable.
	@param r The expression that defines the value of this
	variable.
	@param i Indicates if this is an implicitly defined
	deterministic variable. */
    determVar(const string &n, expr *r, bool i = false);
    
    /// Creates an inherited variable in another record, allocates with new
    virtual variable *createInherited(record *r) const;

    /** Prints the model specification to the standard output.
	@see node::printCode
	@see compiler::printTree */
    void printCode();
   
    /** Finds and links all the variable references to the actual
	variables. This function is needed in the end of the parse
	phase, when all the variable and record references are checked
	through. The function call is repeated recursively to all of
	the variable references in the expression of that defines the
	value of this variable.
	@see compiler::parseInput
	@see variableRef::findReferences */
    inline void findReferences() { rule->findReferences(); }

    /** Collects the variables of the model into the type lists. The
        variable adds itself to the type list corresponding to its C++
	type. The C++ type of a determVar is either composite or
	constant. 
	@see compiler::collectVariables */
    void collectTypes(vector<variable*> *c, vector<variable*> *o,
		      vector<variable*> *);

    /** Generates a C++ expression to calculate the value of this
	variable. */ 
    void genFunction();
    /** Generates the main content of the addDependents function
	implementation, if this is a composite. */ 
    void genDepFunction();
    /** Generates a 'true' into the constructor of this variable, if
	this is an implicit composite. */
    void genImplicit();
    /** Is this an implicit variable? */
    bool isImplicit() { return implicit; }
    
    /// Generates the code that creates a new instance of this variable. 
    void genCreate();
    /** Generates the insertion of this variable into a global map. */
    void genAddToList();
    /// Generates the initialization function call.
    void genInit();
};


#endif // VARIABLE_H

//-----------------------------//
// Local Variables:            //
// mode: c++                   //
// make-backup-files: t        //
// c-basic-offset: 4           //
// End:                        //
//-----------------------------//
